/*
 * spacemit-specific stuff (for k1): soc init, reset
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "../port/error.h"
#include "riscv64.h"

enum {				/* twmer bits */
	We	= 1,
};

typedef struct {
	uchar	_0[0xb0];
	ulong	twfar;
	ulong	twsar;
	ulong	twmer;
} Spacedog;

static void
unlockdog(Spacedog *dog)
{
	dog->twfar = 0xbaba;
	coherence();
	dog->twsar = 0xeb10;
	coherence();
}

static void
spacemitinit(void)
{
	ulong *emac0clkrstctl, *l2ctl;
	Spacedog *dog = (Spacedog *)soc.wdogs[0];

	if(TODO && soc.wdog0) {
		unlockdog(dog);
		dog->twmer = 0;		/* clear We */
		unlockdog(dog);
		*(ulong *)soc.wdog0 = 0; /* watchdog reset + clocks off */
		iprint("watchdog off\n");
	}
	USED(dog);

	/* hasn't helped yet */
	l2ctl  = (ulong *)KADDR(0xd5001000 + 0x1b0);
	l2ctl[0] |= 05;			/* hw enable, hw pmu flush mode */
	l2ctl[1] |= 05;			/* hw enable, hw pmu flush mode */
	coherence();

	/* hasn't helped yet */
	emac0clkrstctl = (ulong *)KADDR(0xd4282800 + 0x3e4);
	*emac0clkrstctl |= 1;		/* enable emac0 axi bus clock */
	coherence();
}

static void
spacemitreset(void)
{
	Spacedog *dog = (Spacedog *)soc.wdogs[0];

	if(soc.wdog0) {
		splhi();
		unlockdog(dog);
		*(ulong *)soc.wdog0 = 7; /* watchdog reset + clocks on */
		unlockdog(dog);
		dog->twmer = We;

		iprint("awaiting watchdog reset\n");
		for (;;)
			halt();
	}
}

void (*socinit)(void) = spacemitinit;
void (*rvreset)(void) = spacemitreset;
