/*
 * driver for Spacemit K1x-emac Gb Ethernet MAC.
 * written without benefit of adequate English documentation.
 * not working on milk-v jupiter yet.
 * can only use bottom 4GB.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/error.h"
#include "../port/netif.h"
#include "etherif.h"

typedef struct Ctlr Ctlr;
typedef struct Rd Rd;
#define Td Rd

enum {
	Rbsz	= ETHERMAXTU+4,	/* with CRC, ≤4k */
	Descalign= 16,		/* a guess */

	/* tunable parameters */
	Ntd	= 64,		/* power of 2 */
	Nrd	= 64,		/* power of 2 */
	Nrb	= 2*Nrd,	/* private receive buffers, could be ≫ Nrd */
};

/* Macaddr and Statdata get padded to 16 bytes.  Declare as array in Regs. */
typedef ulong Macaddrarr[3];
typedef struct Macaddr Macaddr;
struct Macaddr {
	ulong	hi;
	ulong	med;
	ulong	lo;
};

typedef ulong Statdataarr[3];
typedef struct Statdata Statdata;
struct Statdata {
	ulong	ctl;
	ulong	hi;
	ulong	lo;
};

typedef struct Regs Regs;
struct Regs {
	/* dma */
	ulong	dmacfg;
	ulong	dmactl;
	ulong	dmairqsts;
	ulong	dmaintren;

	ulong	dmatxpollctr;
	ulong	dmatxpolldmnd;
	ulong	dmarxpolldmnd;
	ulong	dmatxbase;
	ulong	dmarxbase;

	ulong	dmamissed;
	ulong	dmastopflsh;
	ulong	dmarxirqmitig;

	ulong	dmacurrtxdesc;
	ulong	dmacurrtxbuf;
	ulong	dmacurrrxdesc;
	ulong	dmacurrrxbuf;
	uchar	_0[0x100 - 0x40];

	/* mac */
	ulong	macctl;			/* "global" control */
	ulong	mactxctl;
	ulong	macrxctl;
	ulong	macmaxframe;
	ulong	mactxjabsz;
	ulong	macrxjabsz;
	ulong	macaddrctl;
	ulong	macmdioclkdiv;
	Macaddrarr macaddr[4];
	ulong	macmcasthash[4];

	ulong	macfcctl;
	ulong	macfcpausefrmgen;
	Macaddrarr macfcsrc;
	Macaddrarr macfcdest;
	ulong	macfcpausetm;
	uchar	_1[0x1a0 - 0x184];

	ulong	macmdioctl;
	ulong	macmdiodata;
	Statdataarr macrxstat;
	Statdataarr mactxstat;
	ulong	mactxfifonear;
	ulong	mactxpktstart;
	ulong	macrxpktstart;
	uchar	_2[0x1e0 - 0x1cc];

	ulong	macirqsts;
	ulong	macintren;
};

enum {					/* dmacfg: dma config */
	Swrst	= 1<<0,
	Burstlen= MASK(7)<<1,
	Descskip= MASK(5)<<8,
/* For Receive and Transmit DMA operate in Big-Endian mode for Descriptors. */
	Descbyteorder = 1<<13,
	Descbiglittleend = 1<<14,
	Txrxarb	= 1<<15,
	Waitdone= 1<<16,
	Strictburst = 1<<17,
	Dma64bit= 1<<18,
};

enum {					/* dmactl: dma control */
	Txstart = 1<<0,
	Rxstart = 1<<1,
};

enum {					/* dmairqsts: dma irq status */
	Txdoneirq = 1<<0,
	Txdescunavailirq = 1<<1,
	Txstoppedirq = 1<<2,
	Rxdoneirq = 1<<4,
	Rxdescunavailirq = 1<<5,
	Rxstoppedirq = 1<<6,
	Rxrcvmissedirq = 1<<7,
	Macirq	= 1<<8,
	Txdmastate = MASK(3)<<16,
	Rxdmastate = MASK(4)<<20,
};

enum {					/* dmaintren: dma interrupt enable */
	Txdoneintren = 1<<0,
	Txdescunavailintren = 1<<1,
	Txdmastoppedintren = 1<<2,
	Rxdoneintren = 1<<4,
	Rxdescunavailintren = 1<<5,
	Rxdmastoppedintren = 1<<6,
	Rxmissedintren = 1<<7,
	Macintren = 1<<8,
};

enum {					/* dmarxirqmitig: rx irq status */
	Rxirqfrms = MASK(8),
	Rxirqtmouts = MASK(20)<<8,
	Rxirqfrmmode = 1<<30,
	Rxirqmiten = 1u<<31,
};

enum {					/* macctl: mac control */
	Macspeed	= MASK(2),
	Macspeed1g	= 1<<1,
	Macfulldpx	= 1<<2,
	Macrstrxstats	= 1<<3,
	Macrsttxstats	= 1<<4,
	Macucastwake	= 1<<8,
	Macmagpktwake	= 1<<9,
};

enum {					/* mactxctl: mac transmit control */
	Mactxen	= 1<<0,
	Macinvfcs = 1<<1,
	Macdisfcsins = 1<<2,
	Mactxretry = 1<<3,
	Macifg	= MASK(3)<<4,
	Macpreamb= MASK(3)<<7,
};

enum {					/* macrxctl: mmac receive control */
	Macrxen	= 1<<0,
	Macdisfcsck = 1<<1,
	Macstripfcs = 1<<2,
	Macstorefwd = 1<<3,
	Macsts1st = 1<<4,
	Macaccbad = 1<<5,
	Maccntvlan = 1<<6,
};

enum {					/* macaddrctl: mac address control */
	Macaddr1en = 1<<0,
	Macaddr2en = 1<<1,
	Macaddr3en = 1<<2,
	Macaddr4en = 1<<3,
	Macinvaddr1en = 1<<4,
	Macinvaddr2en = 1<<5,
	Macinvaddr3en = 1<<6,
	Macinvaddr4en = 1<<7,
	Macprom = 1<<8,
};

enum {					/* macmdioctl: mac mdio control */
	Macphyaddr = MASK(5),
	Macregaddr = MASK(5)<<5,
	Macmdiorw = 1<<10,
	Macstmdioxfr = 1<<15,
};

enum {				/* mactxstat.ctl: mac tx stats control */
	Mactxctrno = MASK(5),
	Macsttxctrrd = 1<<15,
};

enum {					/* macirqsts: mac irq status */
	Macunderrunirq = 1<<0,
	Macjabberirq = 1<<1,
};

enum {					/* macintren: mac interrupt enable */
	Macunderrunie = 1<<0,
	Macjabberie = 1<<1,
};

/* this is a register layout, which can't tolerate bogus padding */
struct Rd {				/* receive & transmit descriptor */
	ulong	flags;
	ulong	sizes;
	ulong	buf1;
	ulong	buf2;
};
enum {					/* Rd flags */
	Rdpktlen = MASK(14),
	Rdapplsts = MASK(15)<<14,
	Rdlast	= 1<<29,
	Rdfirst = 1<<30,
	Rdown	= 1<<31,		/* owned by hw (? usual convention) */
};
enum {					/* Rd sizes */
	Rdbufsz1 = MASK(12),
	Rdbufsz2 = MASK(12)<<12,
	Rd2ndchain = 1<<25,
	Rdendring = 1<<26,
};

enum {					/* Td flags */
	Tdpktsts = MASK(30),
	Tdown	= 1<<31,		/* owned by hw (? usual convention) */
};
enum {					/* Td sizes */
	Tdbufsz1 = MASK(12),
	Tdbufsz2 = MASK(12)<<12,
	Td2ndchain = 1<<25,
	Tdendring = 1<<26,

	Tdnopad	= 1<<27,
	Tdnocrc = 1<<28,
	Tdfirst	= 1<<29,
	Tdlast	= 1<<30,
	Tddoneintr= 1<<31,
};

#ifdef unused
struct desc_buf {
	uvlong	dma_addr;
	void	*buff_addr;
	ushort	dma_len;
	uchar	map_as_page;
};
#endif

typedef struct {
	uint	reg;
	char	*name;
} Stat;

static Stat stattab[] = {
	0,	nil,
};

struct Ctlr {
	Ethident;		/* see etherif.h, includes regs* */

	Lock	reglock;
	uint	im;		/* interrupt mask (enables) */
	uint	lim;
	uint	rim;
	uint	tim;

	Rendez	rrendez;
	Rd*	rdba;		/* receive descriptor base address */
	uint	rdh;		/* " " head */
	uint	rdt;		/* " " tail */
	Block**	rb;		/* " buffers */
	uint	rintr;
	uint	rsleep;
	Watermark wmrd;
	Watermark wmrb;
	int	rdfree;		/* " descriptors awaiting packets */

	Rendez	trendez;
	Td*	tdba;		/* transmit descriptor base address */
	uint	tdh;		/* " " head */
	uint	tdt;		/* " " tail */
	Block**	tb;		/* " buffers */
	uint	tintr;
	uint	tsleep;
	Watermark wmtd;
	QLock	tlock;
	ulong	txtick;		/* tick at start of last tx start */

	Rendez	lrendez;

	uchar	flag;
	int	procsrunning;	/* flag: kprocs started for this Ctlr? */
	QLock	alock;		/* attach lock */
	int	attached;

	uchar	ra[Eaddrlen];	/* receive address */

	QLock	slock;
	ulong	stats[nelem(stattab)];

	uint	ixcs;		/* valid hw checksum count */
	uint	ipcs;		/* good hw ip checksums */
	uint	tcpcs;		/* good hw tcp/udp checksums */
};

enum {				/* flag bits */
	Factive	= 1<<0,
};

static	Ctlr	*ctlrtab[8];
static	int	nctlr;

/* these are shared by all the controllers of this type */
static	Lock	rblock;
static	Block	*rbpool;
/* # of rcv Blocks awaiting processing; can be briefly < 0 */
static	int	nrbfull;
static	int	nrbavail;

static void	replenish(Ctlr *ctlr);

static void
readstats(Ctlr *ctlr)
{
	uint i, reg;

	qlock(&ctlr->slock);
	for(i = 0; i < nelem(ctlr->stats); i++) {
		reg = stattab[i].reg;
		ctlr->stats[i] += ctlr->regs[reg / BY2WD];
	}
	qunlock(&ctlr->slock);
}

static long
ifstat(Ether *edev, void *a, long n, ulong offset)
{
	uint i;
	char *s, *p, *e;
	Ctlr *ctlr;
	Regs *regs;

	ctlr = edev->ctlr;
	p = s = malloc(READSTR);
	if(p == nil)
		error(Enomem);
	e = p + READSTR;

	readstats(ctlr);
	for(i = 0; i < nelem(stattab); i++)
		if(stattab[i].name != nil && ctlr->stats[i] > 0)
			p = seprint(p, e, "%s\t%uld\n", stattab[i].name,
				ctlr->stats[i]);
	regs = ctlr->vregs;
	p = seprint(p, e, "%Æ\n", ctlr);
	p = seprint(p, e, "macctl %#lux\n",  regs->macctl);
	p = seprint(p, e, "mactxctl %#lux\n",  regs->mactxctl);
	p = seprint(p, e, "macrxctl %#lux\n",  regs->macrxctl);
	p = seprint(p, e, "dmactl %#lux\n",  regs->dmactl);
	p = seprint(p, e, "dmacfg %#lux\n",  regs->dmacfg);
	p = seprint(p, e, "rintr %d rsleep %d\n", ctlr->rintr, ctlr->rsleep);
	p = seprint(p, e, "tintr %d tsleep %d\n", ctlr->tintr, ctlr->tsleep);
	p = seprintmark(p, e, &ctlr->wmrb);
	p = seprintmark(p, e, &ctlr->wmrd);
	seprintmark(p, e, &ctlr->wmtd);
	n = readstr(offset, a, n, s);
	free(s);

	return n;
}

static void
ienablelcked(Ctlr *ctlr, int ie)
{
	Regs *regs = ctlr->vregs;

	ctlr->im |= ie;
	regs->dmaintren |= ie;
}

static void
ienable(Ctlr *ctlr, int ie)
{
	ilock(&ctlr->reglock);
	ienablelcked(ctlr, ie);
	iunlock(&ctlr->reglock);
}

/* return a Block from our pool */
static Block*
rballoc(void)
{
	Block *bp;

	ilock(&rblock);
	if((bp = rbpool) != nil){
		rbpool = bp->next;
		bp->next = nil;
		adec(&nrbavail);
		if (nrbavail < 0)
			print("etherk1x rballoc: nrbavail negative\n");
	}
	iunlock(&rblock);
	if (bp)
		l2cacheflush(bp->rp, Rbsz);
	return bp;
}

/* called from freeb for receive Blocks, returns them to our pool */
void
rbfree(Block *bp)
{
	bp->wp = bp->rp = (uchar *)ROUNDDN((uintptr)bp->lim - Rbsz, BLOCKALIGN);
	assert(bp->rp >= bp->base);
	assert(((uintptr)bp->rp & (BLOCKALIGN-1)) == 0);
	bp->flag &= ~(Bipck | Budpck | Btcpck | Bpktck);

	ilock(&rblock);
	bp->next = rbpool;
	rbpool = bp;
	adec(&nrbfull);
	ainc(&nrbavail);
	iunlock(&rblock);
}

/* reclaim sent tx buffers & descriptors */
static void
cleanup(Ctlr *ctlr)
{
	uint i, n, tdh;
	Block *bp;
	Td *td;

	tdh = ctlr->tdh;
	i = 0;
	n = NEXT(tdh, Ntd);
	td = &ctlr->tdba[n];
	l2cacheflush(td, sizeof *td);
iprint("C%d ", n);
	while (td = &ctlr->tdba[n], (td->flags & Tdown) == 0 &&
	    n != ctlr->tdt && i++ < Ntd-2) {
		tdh = n;
		bp = ctlr->tb[tdh];
		ctlr->tb[tdh] = nil;
		if (bp)
			freeb(bp);
iprint("tdh %d ", tdh);
		n = NEXT(tdh, Ntd);
		td = &ctlr->tdba[n];
		l2cacheflush(td, sizeof *td);
	}
	ctlr->tdh = tdh;
	notemark(&ctlr->wmrb, nrbfull);
}

void
transmit(Ether *edev)
{
	Block *bp;
	Ctlr *ctlr;
	Regs *regs;
	Td *td;
	uint nqd, tdt, tdh;

	ctlr = edev->ctlr;
	if(!canqlock(&ctlr->tlock)){
		ienable(ctlr, Txdoneintren);
		return;
	}
	if (!ctlr->attached) {
		qunlock(&ctlr->tlock);
		return;
	}
iprint("T");
	cleanup(ctlr);				/* free transmitted buffers */
	tdh = ctlr->tdh;
	tdt = ctlr->tdt;
	assert(ctlr->tdba != nil);
	regs = ctlr->vregs;
	for(nqd = 0; NEXT(tdt, Ntd) != tdh; nqd++){	/* ring not full? */
		td = &ctlr->tdba[tdt];
		l2cacheflush(td, sizeof *td);
		if (td->flags & Tdown)		/* still being xmitted? */
			break;
		if((bp = qget(edev->oq)) == nil)
			break;

		/* populate first available desc. for block bp */
		l2cacheflush(bp->rp, BLEN(bp));
		td->buf1 = PADDR(bp->rp);
		td->buf2 = 0;
		td->sizes &= ~Tdbufsz1;
		td->sizes |= Tdfirst | Tdlast | Tddoneintr | BLEN(bp);
		coherence();
		l2cacheflush(td, sizeof *td);

		if (ctlr->tb[tdt])
			panic("%Æ: xmit q full", ctlr);
		ctlr->tb[tdt] = bp;
		coherence();
iprint("tdt %d ", tdt);

		td->flags |= Tdown;		/* allows xmit */
		l2cacheflush(td, sizeof *td);

		tdt = NEXT(tdt, Ntd);
	}
	/* note size of queue of tds awaiting transmission */
	notemark(&ctlr->wmtd, (uint)(tdt + Ntd - tdh) % Ntd);

	ilock(&ctlr->reglock);
	if(nqd) {
		ctlr->tdt = tdt;
		ctlr->txtick = sys->ticks;
		regs->mactxctl |= Mactxen|Mactxretry;	/* kick transmitter */
		coherence();
		regs->dmatxpolldmnd = 0xff;		/* start dma; magic */
		l2cacheflush(regs, sizeof *regs);
iprint("K%d", tdt);
	}
	ienablelcked(ctlr, Txdoneintren);
	iunlock(&ctlr->reglock);
	qunlock(&ctlr->tlock);
}

static int
tim(void *vc)
{
	Ctlr *ctlr = (void *)vc;

	if(ctlr->tim != 0)
		return 1;
	if(qlen(ctlr->edev->oq) == 0)
		return 0;
	l2cacheflush(&ctlr->tdba[ctlr->tdt], sizeof(Td));
	return (ctlr->tdba[ctlr->tdt].flags & Tdown) == 0;
}

static void
tproc(void *v)
{
	ulong txtick;
	Ctlr *ctlr;
	Ether *edev;

	edev = v;
	ctlr = edev->ctlr;
	for (;;) {
		ctlr->tsleep++;
		/* the time-out is only for diagnosis */
		tsleep(&ctlr->trendez, tim, ctlr, 1000);
		ctlr->tim = 0;
		txtick = ctlr->txtick;
		if (txtick && sys->ticks - txtick > 2*HZ) {
			print("%Æ: not transmitting\n", ctlr);
			ctlr->txtick = 0;
		}
		/*
		 * perhaps some buffers have been transmitted and their
		 * descriptors can be reused to copy Blocks out of edev->oq.
		 */
		transmit(edev);
	}
}

/* free any buffer Blocks in an array */
static void
freeblks(Block **blks, int count)
{
	Block *bp;

	if (blks == nil)
		return;
	while(count-- > 0){
		bp = blks[count];
		blks[count] = nil;
		if (bp) {
			bp->free = nil;
			freeb(bp);
		}
	}
}

static void
initrxring(Ctlr *ctlr, Rd *rdbase, int ndescs)
{
	int i;
	Rd *rd;

	ctlr->rdba = rdbase;
	memset(ctlr->rdba, 0, ndescs * sizeof(Rd));
	for (i = 0; i < ndescs; i++) {
		rd = &ctlr->rdba[i];
		rd->flags = 0;	/* no Rdown prevents rcv filling for now */
		rd->sizes = Rbsz;
	}
	ctlr->rdba[ndescs - 1].sizes |= Rdendring;
	coherence();
	l2cacheflush(ctlr->rdba, ndescs * sizeof(Rd));
}

static void
inittxring(Ctlr *ctlr, Td *tdbase, int ndescs)
{
	int i;
	Td *td;

	ctlr->tdba = tdbase;
	memset(ctlr->tdba, 0, ndescs * sizeof(Td));
	for (i = 0; i < ndescs; i++) {
		td = &ctlr->tdba[i];
		td->flags = 0;		/* available to fill to send */
	}
	ctlr->tdba[ndescs - 1].sizes |= Tdendring;
	coherence();
	l2cacheflush(ctlr->tdba, ndescs * sizeof(Td));
}

/*
 * Get the receiver into a usable state.  Some of this is boilerplate
 * that could be (or is) done automatically as part of reset (hint, hint),
 * but we also disable broken features (e.g., Intel's IP checksums).
 *
 * We're called before txinit, so arrange link auto negotiation here.
 */
static void
rxinit(Ctlr *ctlr)
{
	int i;
	Regs *regs;

	regs = ctlr->vregs;
// dmactl is 0 here, dmacfg is 0x70020 (Waitdone|Strictburst|Dma64bit|16<<1)
	ilock(&ctlr->reglock);
	ctlr->im = 0;
	regs->dmaintren = 0;
	regs->macintren = 0;
	regs->macrxctl &= ~Macrxen;
	regs->mactxctl &= ~Mactxen;
	coherence();
	regs->dmactl = 0;
	coherence();

if (TODO) {
iprint("into reset...");
	regs->dmacfg = Swrst;
	coherence();
	delay(10);
iprint("clearing reset...");
	regs->dmacfg = 0;
	coherence();
	delay(10);
}
	regs->dmacfg &= ~Burstlen;
	regs->dmacfg |= Waitdone|Strictburst|Dma64bit|16<<1;
	regs->macctl = (regs->macctl & ~Macspeed) | Macfulldpx | Macspeed1g;
	regs->macaddrctl = Macaddr1en;
	for (i = 0; i < sizeof regs->macmcasthash; i++)
		regs->macmcasthash[i] = (ushort)~0; /* accept all mcast for now */

	ctlr->rdfree = 0;
	initrxring(ctlr, ctlr->rdba, Nrd);
	regs->dmarxbase = PADDR(ctlr->rdba);
	coherence();

	replenish(ctlr);
	regs->dmactl |= Rxstart;
	regs->macrxctl |= Macrxen;
	coherence();
	regs->macrxpktstart = 12;		/* magic */
	ienablelcked(ctlr, Rxdoneintren | Rxdmastoppedintren | Macintren);
	iunlock(&ctlr->reglock);
}

static void
replenish(Ctlr *ctlr)
{
	uint rdt, rdh;
	Block *bp;
	Block **rb;
	Rd *rd;

	rdh = ctlr->rdh;
	for(rdt = ctlr->rdt; NEXT(rdt, Nrd) != rdh; rdt = NEXT(rdt, Nrd)){
		rd = &ctlr->rdba[rdt];
		if (rd->flags & Rdown)	/* still held by hw? */
			break;
		rb = &ctlr->rb[rdt];
		if(*rb != nil){
			iprint("%Æ: rx overrun\n", ctlr);
			break;
		}
		*rb = bp = rballoc();
		if(bp == nil)		/* don't have a buffer for this desc? */
			break;

		/* populate first available desc. for new block bp */
		rd->buf1 = PADDR(bp->rp);
		rd->buf2 = 0;
		rd->sizes &= ~Rdbufsz1;
		rd->sizes |= Rbsz;
		coherence();
		l2cacheflush(rd, sizeof *rd);
		rd->flags &= ~Rdpktlen;
		rd->flags |= Rdown | Rbsz;	/* Rdown allows reception */
		coherence();
		l2cacheflush(rd, sizeof *rd);
		ctlr->rdfree++;
	}
	ctlr->rdt = rdt;
}

static int
rim(void *vc)
{
	Ctlr *ctlr = (void *)vc;

	if(ctlr->rim != 0)
		return 1;
	l2cacheflush(&ctlr->rdba[ctlr->rdh], sizeof(Rd));
	return (ctlr->rdba[ctlr->rdh].flags & Rdown) == 0;
}

static void
ckcksum(Ctlr *ctlr, Rd *, Block *bp)
{
	ctlr->ixcs++;
	bp->flag |= Bpktck;
}

//enum {
//	Rxdonebits = Rxoverrun | Rxbufnotavail | Rxrcvd,
//};

static int
qinpkt(Ctlr *ctlr)
{
	int passed;
	uint rdh, len;
	Block *bp;
	Rd *rd;

	ctlr->rim = 0;
	rdh = ctlr->rdh;
	rd = &ctlr->rdba[rdh];
	l2cacheflush(rd, sizeof *rd);
	if (rd->flags & Rdown)		/* still held by hw? */
		return -1;		/* wait for pkts to arrive */

	/*
	 * Accept packets with no errors.
	 */
	passed = 0;
	coherence();
	bp = ctlr->rb[rdh];
	if ((rd->flags & (Rdlast|Rdfirst)) == (Rdlast|Rdfirst)){
		if (bp == nil) {
			print("%Æ: nil Block* from ctlr->rb\n", ctlr);
			return -1;
		}
		len = rd->flags & Rdpktlen;
		if (len == ETHERMINTU)
			len += 4;	/* for 60-byte arps? */
		if (len <= ETHERMAXTU)
			bp->wp += len;
		ckcksum(ctlr, rd, bp);
		/* note size of queue of Blocks awaiting transmission */
		notemark(&ctlr->wmrb, ainc(&nrbfull));
		coherence();
		/* pass pkt upstream, where it will be freed eventually */
		l2cacheflush(bp->rp, len);
		etheriq(ctlr->edev, bp, 1);
		passed++;
	} else {
		ainc(&nrbfull);
		freeb(bp);			/* toss bad pkt */
		ilock(&ctlr->reglock);
//		ctlr->regs[Rxsts] |= Rxdonebits; /* reset done bits */
		iunlock(&ctlr->reglock);
	}

	ctlr->rb[rdh] = nil;
	ctlr->rdh = NEXT(rdh, Nrd);
	ctlr->rdfree--;
	/*
	 * if number of rds ready for packets is too low,
	 * set up the unready ones.
	 */
	if (ctlr->rdfree <= Nrd*3/4)
		replenish(ctlr);
	return passed;
}

static void
rproc(void *v)
{
	int passed, npass;
	Ctlr *ctlr;
	Ether *edev;

	edev = v;
	ctlr = edev->ctlr;
	for (ctlr->rdh = 0; ; ) {
		replenish(ctlr);
		/*
		 * Prevent an idle or unplugged interface from interrupting.
		 * Allow receiver interrupts initially and again
		 * if the interface (and transmitter) see actual use.
		 */
		if (edev->outpackets > 10 || ctlr->rintr < 2*Nrd)
			ienable(ctlr, Rxdoneintren);
		ctlr->rsleep++;
		sleep(&ctlr->rrendez, rim, ctlr);

		for(passed = 0; (npass = qinpkt(ctlr)) >= 0; passed += npass)
			;
		/* note how many rds had full buffers */
		notemark(&ctlr->wmrd, passed);
	}
}

static void
promiscuous(void *a, int on)
{
	Ctlr *ctlr;
	Ether *edev;
	Regs *regs;

	edev = a;
	ctlr = edev->ctlr;
	regs = ctlr->vregs;
	ilock(&ctlr->reglock);
	if(on)
		regs->macaddrctl |= Macprom;
	else
		regs->macaddrctl &= ~Macprom;
	iunlock(&ctlr->reglock);
}

/* TODO: find right algorithm for k1x; see §16.2.3 of the Xilinx Zynq-7000 SoC Tech. Ref. Man. */
static ulong
mcasthash(uchar *mac)
{
	uint chr, bit, hash;

	hash = 0;
	for (chr = 0; chr < Eaddrlen; chr++)
		for (bit = chr; bit < 8*Eaddrlen; bit += 6)
			if (mac[bit>>3] & (1 << (bit & 7)))
				hash ^= 1 << chr;
	return hash;
}

static void
multicast(void *a, uchar *addr, int on)
{
	ulong hash, word, bit;
	Ctlr *ctlr;
	Ether *edev;
	Regs *regs;

	edev = a;
	if (edev == nil)
		panic("multicast: nil edev arg");
	ctlr = edev->ctlr;
	if (ctlr == nil)
		panic("multicast: nil edev->ctlr");

	/*
	 * multiple ether addresses can hash to the same filter bit,
	 * so it's never safe to clear a filter bit.
	 * if we want to clear filter bits, we need to keep track of
	 * all the multicast addresses in use, clear all the filter bits,
	 * then set the ones corresponding to in-use addresses.
	 */
	SET(regs); USED(regs);
	if (1)			// TODO: use right multicast hash algorithm
		return;
	regs = ctlr->vregs;
	USED(regs);
	hash = mcasthash(addr);
	word = BITMAPWD(hash);
	bit = BITMAPBIT(hash);
	ilock(&ctlr->reglock);
	if(on)
		ctlr->regs[word] |= bit;
//	else
//		ctlr->regs[word] &= ~bit;
	iunlock(&ctlr->reglock);
}

static void
freez(void **pptr)
{
	free(*pptr);
	*pptr = nil;
}

static void
freemem(Ctlr *ctlr)
{
	int i;
	Block *bp;

	/* only free enough rcv bufs for one controller */
	for (i = Nrb; i > 0 && (bp = rballoc()) != nil; i--){
		bp->free = nil;
		freeb(bp);
	}
	freez(&ctlr->rdba);
	freez(&ctlr->tdba);
	freez(&ctlr->rb);
	freez(&ctlr->tb);
}

/* don't discard all state; we may be attached again */
static int
detach(Ctlr *ctlr)
{
	Regs *regs;

	regs = ctlr->vregs;
	ilock(&ctlr->reglock);
	ctlr->im = 0;
	regs->dmaintren &= ~(Txdoneintren | Rxdoneintren | Macintren);
//	regs->macintren &= ~(Txdoneintren | Rxdoneintren | Macintren);
	regs->mactxctl &= ~Mactxen;
	regs->macrxctl &= ~Macrxen;
//	regs->macctl = 0;
	iunlock(&ctlr->reglock);
	delay(10);

	ctlr->attached = 0;
	return 0;
}

static void
shutdown(Ether *edev)
{
	detach(edev->ctlr);
	/* don't freemem; we may reattach */
}

uchar ether0mac[];

static void
setmacs(Ctlr *ctlr)
{
	ulong top, bot;
	uchar *ra;
	Macaddr *mp;
	Regs *regs;
	static int ctlrno;

	regs = ctlr->vregs;
	ra = ctlr->ra;
	ilock(&ctlr->reglock);
	mp = (Macaddr *)regs->macaddr[0];
	bot = mp->med << 16 | mp->hi;
	top = mp->lo;
	if (bot || top) {
		ra[0] = bot;
		ra[1] = bot>>8;
		ra[2] = bot>>16;
		ra[3] = bot>>24;
		ra[4] = top;
		ra[5] = top>>8;
	} else {
		iprint("#l0: mac unset; setting to %E\n", ether0mac);
		memmove(ra, ether0mac, sizeof ra);
		ra[5] += ctlrno++;		/* big endian mac */
		mp->lo  = ra[5]<<8 | ra[4];
		mp->med = ra[3]<<8 | ra[2];
		mp->hi  = ra[1]<<8 | ra[0];
		coherence();
	}
	regs->macaddrctl = Macaddr1en;
	iunlock(&ctlr->reglock);
	if (ctlr->edev)
		multicast(ctlr->edev, ra, 1);
}

/* may be called from discover with ctlr only partially populated */
static int
reset(Ctlr *ctlr)
{
	uint *regs;

	iprint("resetting etherk1x, we hope...");
	if(detach(ctlr)){
		print("%Æ: reset timeout\n", ctlr);
		return -1;
	}

	regs = ctlr->regs;
	if (regs == nil) {
		print("%Æ: nil regs\n", ctlr);
		return -1;
	}
	/* if unknown, load mac address from non-volatile memory, if present */
	setmacs(ctlr);
	readstats(ctlr);		/* zero stats by reading regs */
	memset(ctlr->stats, 0, sizeof ctlr->stats);
	iprint("\n");
	return 0;
}

/*
 * Get the transmitter into a usable state.  Much of this is boilerplate
 * that could be (or is) done automatically as part of reset (hint, hint).
 */
static void
txinit(Ctlr *ctlr)
{
	Regs *regs;

	regs = ctlr->vregs;
	ilock(&ctlr->reglock);
	regs->dmactl &= ~Txstart;
	coherence();
	regs->mactxctl &= ~Mactxen;
	coherence();
	delay(10);

	/* set up tx queue 0 ring */
	inittxring(ctlr, ctlr->tdba, Ntd);
	ctlr->tdh = Ntd - 1;
	regs->dmatxbase = PADDR(ctlr->tdba);
	regs->dmatxpolldmnd = 0;		/* magic */
	coherence();

	regs->mactxfifonear = 504;
	regs->mactxpktstart = 192;		/* magic */
	coherence();
	regs->mactxctl |= Mactxen|Mactxretry;	/* kick transmitter */
	coherence();
	regs->dmactl |= Txstart;
	iunlock(&ctlr->reglock);
}

static void
allocall(Ctlr *ctlr)
{
	int i;
	Block *bp;

	/* discard any buffer Blocks left over from before detach */
	freeblks(ctlr->tb, Ntd);
	freeblks(ctlr->rb, Nrd);

	if (ctlr->rdba == nil)
		ctlr->rdba = mallocalign(Nrd * sizeof(Rd), Descalign, 0, 0);
	if (ctlr->tdba == nil)
		ctlr->tdba = mallocalign(Ntd * sizeof(Td), Descalign, 0, 0);
	if (ctlr->rdba == nil || ctlr->tdba == nil)
		error(Enomem);

	ctlr->rb = malloc(Nrd * sizeof(Block *));
	ctlr->tb = malloc(Ntd * sizeof(Block *));
	if (ctlr->rb == nil || ctlr->tb == nil)
		error(Enomem);

	/* add enough rcv bufs for one controller to the pool */
	for(i = 0; i < Nrb; i++){
		bp = allocb(Rbsz);
		if(bp == nil)
			error(Enomem);
		bp->free = rbfree;
		freeb(bp);
	}
	aadd(&nrbfull, Nrb);		/* compensate for adecs in rbfree */
}

static void
etherkproc(Ether *edev, void (*kp)(void *), char *sfx)
{
	char buf[KNAMELEN];

	snprint(buf, sizeof buf, "#l%d%s", edev->ctlrno, sfx);
	kproc(buf, kp, edev);
}

static void
startkprocs(Ctlr *ctlr)
{
	Ether *edev;

	if (ctlr->procsrunning)
		return;
	edev = ctlr->edev;
	/* no obvious way to check link state yet; needs mii/phy goo */
	edev->link = 1;
	etherkproc(edev, tproc, "xmit");
	etherkproc(edev, rproc, "rcv");
	ctlr->procsrunning = 1;
}

static void
wrvfy(uint *reg, uint val)
{
	uint rdval;

	*reg = val;
	coherence();
	rdval = *reg;
	if (rdval != val)
		print("wrote %#ux to %#p, but read back %#ux\n",
			val, reg, rdval);
}

/* could be attaching again after a detach */
static void
attach(Ether *edev)
{
	Ctlr *ctlr;

	ctlr = edev->ctlr;
	qlock(&ctlr->alock);
	if (ctlr->attached) {
		qunlock(&ctlr->alock);
		return;
	}
	if(waserror()){
		reset(ctlr);
		freemem(ctlr);
		qunlock(&ctlr->alock);
		nexterror();
	}
	if(ctlr->rdba == nil)
		allocall(ctlr);
	/* don't change nrbfull; it's shared by all controllers */
	initmark(&ctlr->wmrb, Nrb, "rcv Blocks not yet processed");
	initmark(&ctlr->wmrd, Nrd-1, "rcv descrs processed at once");
	initmark(&ctlr->wmtd, Ntd-1, "xmit descr queue len");
	rxinit(ctlr);
	txinit(ctlr);
	startkprocs(ctlr);
	delay(1000);	/* let phy auto-negotiate; TODO: doesn't help */
	ctlr->attached = 1;
	qunlock(&ctlr->alock);
	poperror();
}

static Intrsvcret
interrupt(Ureg*, void *v)
{
	int icr, icrreg, im, waker, waket, wakel;
	Regs *regs;
	Ctlr *ctlr;

	/* mtrap and strap start with a fence */
	ctlr = ((Ether *)v)->ctlr;
	regs = ctlr->vregs;
	ilock(&ctlr->reglock);
	icrreg = regs->dmairqsts;
	regs->dmaintren = 0;		/* disable all my intrs */
	coherence();
	im = ctlr->im;
	icr = icrreg & im;
	if (0 && icr == 0) {		/* mainly for debugging */
		regs->dmaintren |= im;
		regs->dmairqsts = icrreg;	/* extinguish interrupts */
		iunlock(&ctlr->reglock);
		if (icrreg != 0)
			iprint("%Æ: uninteresting intr, icr %#ux vs im %#ux\n",
				ctlr, icrreg, im);
		return Intrnotforme;
	}

	waker = waket = wakel = 0;
	if(icr & Rxdoneirq){
		ctlr->rim = Rxdoneirq;
		regs->dmaintren &= ~Rxdoneintren;	/* reset */
		waker = 1;
		im &= ~Rxdoneirq;
		ctlr->rintr++;
	}
	if(icr & Txdoneirq){
		ctlr->tim = Txdoneirq;
		regs->dmaintren &= ~Txdoneintren;	/* reset */
		ctlr->txtick = 0;
		waket = 1;
		im &= ~Txdoneirq;
		ctlr->tintr++;
	}
#ifdef TODO_link
	if(icr & Linkchange){
		ctlr->lim = Linkchange;
		regs[Intsts] |= Linkchange;	/* reset */
		wakel = 1;
		/* im &= ~Linkchange;	/* we always want to see link changes */
	}
#endif
	regs->dmairqsts = icrreg;	/* extinguish interrupts */
	/* enable only intr sources we didn't service and are interested in */
	regs->dmaintren = ctlr->im = im;
	iunlock(&ctlr->reglock);

	/* now that registers have been updated, wake sleepers */
	if(waker)
		wakeup(&ctlr->rrendez);
	if(waket)
		wakeup(&ctlr->trendez);
	if(wakel)
		wakeup(&ctlr->lrendez);
	return Intrforme;
}

/*
 * map device p and populate a new Ctlr for it.
 * add the new Ctlr to our list.
 */
static Ctlr *
newctlr(void *regs)
{
	uintptr io;
	void *mem;
	Ctlr *ctlr;

	if (regs == nil)
		return nil;
	io = (uintptr)regs;
	mem = vmap(io, PGSZ);
	if(mem == nil){
		print("#l%d: etherk1x: can't map regs %#p\n", nctlr, regs);
		return nil;
	}
	if(probeulong(mem, Read) < 0) {
		vunmap(mem, PGSZ);
		return nil;
	}

	ctlr = malloc(sizeof *ctlr);
	if(ctlr == nil) {
		vunmap(mem, PGSZ);
		error(Enomem);
	}
	ctlr->regs = (uint*)mem;
	ctlr->physreg = (uint*)io;
	ctlr->prtype = "k1x";
	if(reset(ctlr)){
		print("%Æ: can't reset\n", ctlr);
		free(ctlr);
		vunmap(mem, PGSZ);
		return nil;
	}
	ctlrtab[nctlr++] = ctlr;
	return ctlr;
}

#include <ip.h>

static void
discover(void)
{
	fmtinstall(L'Æ', etherfmt);

	fmtinstall('i', eipfmt);		/* paranoia */
	fmtinstall('I', eipfmt);
	fmtinstall('E', eipfmt);
	fmtinstall('V', eipfmt);
	fmtinstall('M', eipfmt);

	newctlr(soc.ether[0]);
//	newctlr(soc.ether[1]);	/* faults; apparently not enabled by u-boot */
}

static int
addk1x(Ether *edev)
{
	int i;
	Ctlr *ctlr;
	static int ctlrno;

	if (edev == nil)
		panic("etherk1x pnp: nil edev arg");
	if(nctlr == 0)
		discover();
	ctlr = nil;
	for(i = 0; i < nctlr; i++){
		ctlr = ctlrtab[i];
		if(ctlr == nil || ctlr->flag & Factive)
			continue;
		if(edev->port == 0 || edev->port == (uintptr)ctlr->physreg)
			break;
	}
	if (i >= nctlr || ctlr == nil)
		return -1;
	ctlr->flag |= Factive;

	ctlr->edev = edev;		/* point back to Ether */
	edev->ctlr = ctlr;
	edev->port = (uintptr)ctlr->physreg;	/* for printing */
	/* TODO: .irqs could be >1 */
	edev->irq = ioconf("ether", ctlrno++)->irq;
	edev->pcidev = nil;
	edev->tbdf = BUSUNKNOWN;
	edev->mbps = 1000;
	edev->maxmtu = ETHERMAXTU;
	memmove(edev->ea, ctlr->ra, Eaddrlen);

	edev->arg = edev;
	edev->attach = attach;
	edev->detach = shutdown;
	edev->transmit = transmit;
	edev->interrupt = interrupt;
	edev->ifstat = ifstat;
	edev->shutdown = shutdown;
	edev->ctl = nil;
	edev->multicast = multicast;
	edev->promiscuous = promiscuous;
	return 0;
}

void
etherk1xlink(void)
{
	addethercard("k1x", addk1x);
}
