#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "../port/error.h"

#include "io.h"

extern Dev rootdevtab;
extern Dev consdevtab;
extern Dev archdevtab;
extern Dev envdevtab;
extern Dev pipedevtab;
extern Dev procdevtab;
extern Dev mntdevtab;
extern Dev srvdevtab;
extern Dev dupdevtab;
extern Dev ssldevtab;
extern Dev tlsdevtab;
extern Dev capdevtab;
extern Dev kprofdevtab;
extern Dev segmentdevtab;
extern Dev etherdevtab;
extern Dev ipdevtab;
extern Dev uartdevtab;
extern Dev bridgedevtab;
Dev* devtab[] = {
	&rootdevtab,
	&consdevtab,
	&archdevtab,
	&envdevtab,
	&pipedevtab,
	&procdevtab,
	&mntdevtab,
	&srvdevtab,
	&dupdevtab,
	&ssldevtab,
	&tlsdevtab,
	&capdevtab,
	&kprofdevtab,
	&segmentdevtab,
	&etherdevtab,
	&ipdevtab,
	&uartdevtab,
	&bridgedevtab,
	nil,
};

int
acpiinit(void)
{
	return -1;
}

extern uchar boot_CONF_outcode[];
extern usize boot_CONF_outlen;
extern uchar _riscv64_bin_auth_factotumcode[];
extern usize _riscv64_bin_auth_factotumlen;
extern uchar _riscv64_bin_ip_ipconfigcode[];
extern usize _riscv64_bin_ip_ipconfiglen;
extern uchar _cfg_example_nvram_collcode[];
extern usize _cfg_example_nvram_colllen;
extern void ethertemulink(void);
extern void ethermediumlink(void);
extern void loopbackmediumlink(void);
extern void netdevmediumlink(void);
void
links(void)
{
	addbootfile("boot", boot_CONF_outcode, boot_CONF_outlen);
	addbootfile("factotum", _riscv64_bin_auth_factotumcode, _riscv64_bin_auth_factotumlen);
	addbootfile("ipconfig", _riscv64_bin_ip_ipconfigcode, _riscv64_bin_ip_ipconfiglen);
	addbootfile("nvram", _cfg_example_nvram_collcode, _cfg_example_nvram_colllen);
	ethertemulink();
	ethermediumlink();
	loopbackmediumlink();
	netdevmediumlink();
}

#include "../ip/ip.h"
extern void tcpinit(Fs*);
extern void udpinit(Fs*);
extern void ipifcinit(Fs*);
extern void icmpinit(Fs*);
extern void icmp6init(Fs*);
void (*ipprotoinit[])(Fs*) = {
	tcpinit,
	udpinit,
	ipifcinit,
	icmpinit,
	icmp6init,
	nil,
};

extern PhysUart i8250physuart;
PhysUart* physuart[] = {
	&i8250physuart,
	nil,
};

Physseg physseg[8] = {
	{	.attr	= SG_SHARED,
		.name	= "shared",
		.size	= SEGMAXSIZE,
	},
	{	.attr	= SG_BSS,
		.name	= "memory",
		.size	= SEGMAXSIZE,
	},
};
int nphysseg = 8;

char dbgflg[256];

void
rdb(void)
{
	splhi();
	iprint("rdb...not installed\n");
	for(;;);
}

#include "riscv64.h"
int cpuserver = 1;
uvlong cpuhz = 34*1000*1000;  /* from timesync, emulated on 3ghz nuc */
uvlong timebase = 10*1000*1000;	/* clint ticks per second */
Membank membanks[] = {	 	/* (address, size) pairs */
PHYSMEM, BANK0SIZE,
0
};
char defnvram[] = "/boot/nvram";
uintptr uartregs[] = { PAUart0 };
int nuart = nelem(uartregs);
vlong uartfreq = 384000;
uchar ether0mac[] = { 2, 0, 0, 0, 0, 1 };
/* the emulated plic doesn't seem to follow the spec; we ignore it. */
Soc soc = {
.clint	= (char *)PAClint,
.uart	= (char *)PAUart0,
.plic	= (char *)0x40100000,
.ether[0] = (char *)0x40011000,
.hobbled = 0,		/* only 1 hart */
};
Ioconf socconf[] = { /* devices without drivers that vmap their regs */
{ "clint", 64*KB, &soc.clint, },
{ "uart",  PGSZ, &soc.uart, 1, },
{ "plic",  4*MB, &soc.plic, },		/* common but smaller */
0
};
Ioconf ioconfs[] = {	/* devices whose drivers vmap their regs */
{ "ether", 2*PGSZ, &soc.ether[0], 2, },
0
};

char* conffile = "/sys/src/9k/rv/tecpu";
ulong kerndate = KERNDATE;
char *kerndatestr = DATESTR;
char configfile[]={
0x23,0x20,0x39,0x6b,0x20,0x72,0x69,0x73,0x63,0x2d,0x76,0x20,0x63,0x70,0x75,0x20,
0x73,0x65,0x72,0x76,0x65,0x72,0x20,0x66,0x6f,0x72,0x20,0x74,0x69,0x6e,0x79,0x65,
0x6d,0x75,0x0a,0x64,0x65,0x76,0x20,0x2b,0x64,0x65,0x76,0x0a,0x09,0x72,0x6f,0x6f,
0x74,0x0a,0x09,0x63,0x6f,0x6e,0x73,0x0a,0x09,0x61,0x72,0x63,0x68,0x0a,0x09,0x65,
0x6e,0x76,0x0a,0x09,0x70,0x69,0x70,0x65,0x0a,0x09,0x70,0x72,0x6f,0x63,0x0a,0x09,
0x6d,0x6e,0x74,0x0a,0x09,0x73,0x72,0x76,0x0a,0x09,0x64,0x75,0x70,0x0a,0x23,0x09,
0x72,0x74,0x63,0x0a,0x09,0x73,0x73,0x6c,0x0a,0x09,0x74,0x6c,0x73,0x0a,0x09,0x63,
0x61,0x70,0x0a,0x09,0x6b,0x70,0x72,0x6f,0x66,0x0a,0x09,0x73,0x65,0x67,0x6d,0x65,
0x6e,0x74,0x0a,0x09,0x65,0x74,0x68,0x65,0x72,0x09,0x09,0x6e,0x65,0x74,0x69,0x66,
0x0a,0x09,0x69,0x70,0x09,0x09,0x61,0x72,0x70,0x20,0x63,0x68,0x61,0x6e,0x64,0x69,
0x61,0x6c,0x20,0x69,0x70,0x20,0x69,0x70,0x76,0x36,0x20,0x69,0x70,0x61,0x75,0x78,
0x20,0x69,0x70,0x72,0x6f,0x75,0x74,0x65,0x20,0x6e,0x65,0x74,0x6c,0x6f,0x67,0x20,
0x6e,0x75,0x6c,0x6c,0x6d,0x65,0x64,0x69,0x75,0x6d,0x20,0x70,0x6b,0x74,0x6d,0x65,
0x64,0x69,0x75,0x6d,0x0a,0x09,0x75,0x61,0x72,0x74,0x0a,0x09,0x62,0x72,0x69,0x64,
0x67,0x65,0x09,0x09,0x6c,0x6f,0x67,0x0a,0x23,0x09,0x64,0x72,0x61,0x77,0x09,0x73,
0x63,0x72,0x65,0x65,0x6e,0x0a,0x0a,0x75,0x61,0x72,0x74,0x20,0x2b,0x64,0x65,0x76,
0x0a,0x09,0x75,0x61,0x72,0x74,0x69,0x38,0x32,0x35,0x30,0x0a,0x0a,0x69,0x70,0x20,
0x2b,0x64,0x65,0x76,0x0a,0x09,0x74,0x63,0x70,0x0a,0x09,0x75,0x64,0x70,0x0a,0x09,
0x69,0x70,0x69,0x66,0x63,0x0a,0x09,0x69,0x63,0x6d,0x70,0x0a,0x09,0x69,0x63,0x6d,
0x70,0x36,0x0a,0x23,0x09,0x69,0x70,0x6d,0x75,0x78,0x0a,0x0a,0x6c,0x69,0x6e,0x6b,
0x20,0x2b,0x64,0x65,0x76,0x0a,0x09,0x65,0x74,0x68,0x65,0x72,0x74,0x65,0x6d,0x75,
0x0a,0x09,0x65,0x74,0x68,0x65,0x72,0x6d,0x65,0x64,0x69,0x75,0x6d,0x0a,0x09,0x6c,
0x6f,0x6f,0x70,0x62,0x61,0x63,0x6b,0x6d,0x65,0x64,0x69,0x75,0x6d,0x0a,0x09,0x6e,
0x65,0x74,0x64,0x65,0x76,0x6d,0x65,0x64,0x69,0x75,0x6d,0x0a,0x23,0x09,0x75,0x73,
0x62,0x65,0x68,0x63,0x69,0x09,0x09,0x75,0x73,0x62,0x65,0x68,0x63,0x69,0x70,0x63,
0x0a,0x0a,0x6d,0x69,0x73,0x63,0x20,0x2b,0x64,0x65,0x76,0x0a,0x09,0x63,0x61,0x63,
0x68,0x65,0x0a,0x0a,0x23,0x0a,0x23,0x62,0x6f,0x6f,0x74,0x20,0x63,0x70,0x75,0x0a,
0x23,0x09,0x69,0x6e,0x74,0x20,0x63,0x70,0x75,0x66,0x6c,0x61,0x67,0x20,0x3d,0x20,
0x31,0x3b,0x0a,0x23,0x62,0x6f,0x6f,0x74,0x20,0x63,0x70,0x75,0x20,0x62,0x6f,0x6f,
0x74,0x20,0x24,0x33,0x0a,0x23,0x09,0x69,0x6e,0x74,0x20,0x63,0x70,0x75,0x66,0x6c,
0x61,0x67,0x20,0x3d,0x20,0x31,0x3b,0x0a,0x23,0x09,0x63,0x68,0x61,0x72,0x2a,0x20,
0x62,0x6f,0x6f,0x74,0x64,0x69,0x73,0x6b,0x20,0x3d,0x20,0x22,0x24,0x33,0x22,0x3b,
0x0a,0x23,0x62,0x6f,0x6f,0x74,0x20,0x72,0x6f,0x6f,0x74,0x64,0x69,0x72,0x20,0x24,
0x33,0x0a,0x23,0x09,0x63,0x68,0x61,0x72,0x2a,0x20,0x72,0x6f,0x6f,0x74,0x64,0x69,
0x72,0x20,0x3d,0x20,0x22,0x24,0x33,0x22,0x3b,0x0a,0x23,0x62,0x6f,0x6f,0x74,0x20,
0x28,0x62,0x62,0x6f,0x6f,0x74,0x7c,0x72,0x6f,0x6d,0x62,0x6f,0x6f,0x74,0x7c,0x64,
0x6f,0x73,0x62,0x6f,0x6f,0x74,0x29,0x0a,0x23,0x09,0x69,0x6e,0x74,0x20,0x63,0x70,
0x75,0x66,0x6c,0x61,0x67,0x20,0x3d,0x20,0x31,0x3b,0x0a,0x23,0x09,0x63,0x68,0x61,
0x72,0x2a,0x20,0x62,0x6f,0x6f,0x74,0x70,0x72,0x6f,0x67,0x20,0x3d,0x20,0x24,0x32,
0x3b,0x0a,0x23,0x62,0x6f,0x6f,0x74,0x20,0x62,0x6f,0x6f,0x74,0x20,0x24,0x33,0x0a,
0x23,0x09,0x63,0x68,0x61,0x72,0x2a,0x20,0x62,0x6f,0x6f,0x74,0x64,0x69,0x73,0x6b,
0x20,0x3d,0x20,0x22,0x24,0x33,0x22,0x3b,0x0a,0x23,0x0a,0x62,0x6f,0x6f,0x74,0x20,
0x63,0x70,0x75,0x0a,0x09,0x74,0x63,0x70,0x0a,0x0a,0x72,0x6f,0x6f,0x74,0x64,0x69,
0x72,0x0a,0x09,0x62,0x6f,0x6f,0x74,0x24,0x43,0x4f,0x4e,0x46,0x2e,0x6f,0x75,0x74,
0x09,0x09,0x62,0x6f,0x6f,0x74,0x0a,0x09,0x2f,0x72,0x69,0x73,0x63,0x76,0x36,0x34,
0x2f,0x62,0x69,0x6e,0x2f,0x61,0x75,0x74,0x68,0x2f,0x66,0x61,0x63,0x74,0x6f,0x74,
0x75,0x6d,0x0a,0x09,0x2f,0x72,0x69,0x73,0x63,0x76,0x36,0x34,0x2f,0x62,0x69,0x6e,
0x2f,0x69,0x70,0x2f,0x69,0x70,0x63,0x6f,0x6e,0x66,0x69,0x67,0x0a,0x09,0x2f,0x63,
0x66,0x67,0x2f,0x65,0x78,0x61,0x6d,0x70,0x6c,0x65,0x2f,0x6e,0x76,0x72,0x61,0x6d,
0x2e,0x63,0x6f,0x6c,0x6c,0x09,0x6e,0x76,0x72,0x61,0x6d,0x0a,0x0a,0x63,0x6f,0x6e,
0x66,0x0a,0x09,0x23,0x69,0x6e,0x63,0x6c,0x75,0x64,0x65,0x20,0x22,0x72,0x69,0x73,
0x63,0x76,0x36,0x34,0x2e,0x68,0x22,0x0a,0x0a,0x09,0x69,0x6e,0x74,0x20,0x63,0x70,
0x75,0x73,0x65,0x72,0x76,0x65,0x72,0x20,0x3d,0x20,0x31,0x3b,0x0a,0x09,0x75,0x76,
0x6c,0x6f,0x6e,0x67,0x20,0x63,0x70,0x75,0x68,0x7a,0x20,0x3d,0x20,0x33,0x34,0x2a,
0x31,0x30,0x30,0x30,0x2a,0x31,0x30,0x30,0x30,0x3b,0x20,0x20,0x2f,0x2a,0x20,0x66,
0x72,0x6f,0x6d,0x20,0x74,0x69,0x6d,0x65,0x73,0x79,0x6e,0x63,0x2c,0x20,0x65,0x6d,
0x75,0x6c,0x61,0x74,0x65,0x64,0x20,0x6f,0x6e,0x20,0x33,0x67,0x68,0x7a,0x20,0x6e,
0x75,0x63,0x20,0x2a,0x2f,0x0a,0x09,0x75,0x76,0x6c,0x6f,0x6e,0x67,0x20,0x74,0x69,
0x6d,0x65,0x62,0x61,0x73,0x65,0x20,0x3d,0x20,0x31,0x30,0x2a,0x31,0x30,0x30,0x30,
0x2a,0x31,0x30,0x30,0x30,0x3b,0x09,0x2f,0x2a,0x20,0x63,0x6c,0x69,0x6e,0x74,0x20,
0x74,0x69,0x63,0x6b,0x73,0x20,0x70,0x65,0x72,0x20,0x73,0x65,0x63,0x6f,0x6e,0x64,
0x20,0x2a,0x2f,0x0a,0x09,0x4d,0x65,0x6d,0x62,0x61,0x6e,0x6b,0x20,0x6d,0x65,0x6d,
0x62,0x61,0x6e,0x6b,0x73,0x5b,0x5d,0x20,0x3d,0x20,0x7b,0x09,0x20,0x09,0x2f,0x2a,
0x20,0x28,0x61,0x64,0x64,0x72,0x65,0x73,0x73,0x2c,0x20,0x73,0x69,0x7a,0x65,0x29,
0x20,0x70,0x61,0x69,0x72,0x73,0x20,0x2a,0x2f,0x0a,0x09,0x09,0x50,0x48,0x59,0x53,
0x4d,0x45,0x4d,0x2c,0x20,0x42,0x41,0x4e,0x4b,0x30,0x53,0x49,0x5a,0x45,0x2c,0x0a,
0x09,0x09,0x30,0x0a,0x09,0x7d,0x3b,0x0a,0x09,0x63,0x68,0x61,0x72,0x20,0x64,0x65,
0x66,0x6e,0x76,0x72,0x61,0x6d,0x5b,0x5d,0x20,0x3d,0x20,0x22,0x2f,0x62,0x6f,0x6f,
0x74,0x2f,0x6e,0x76,0x72,0x61,0x6d,0x22,0x3b,0x0a,0x0a,0x09,0x75,0x69,0x6e,0x74,
0x70,0x74,0x72,0x20,0x75,0x61,0x72,0x74,0x72,0x65,0x67,0x73,0x5b,0x5d,0x20,0x3d,
0x20,0x7b,0x20,0x50,0x41,0x55,0x61,0x72,0x74,0x30,0x20,0x7d,0x3b,0x0a,0x09,0x69,
0x6e,0x74,0x20,0x6e,0x75,0x61,0x72,0x74,0x20,0x3d,0x20,0x6e,0x65,0x6c,0x65,0x6d,
0x28,0x75,0x61,0x72,0x74,0x72,0x65,0x67,0x73,0x29,0x3b,0x0a,0x09,0x76,0x6c,0x6f,
0x6e,0x67,0x20,0x75,0x61,0x72,0x74,0x66,0x72,0x65,0x71,0x20,0x3d,0x20,0x33,0x38,
0x34,0x30,0x30,0x30,0x3b,0x0a,0x0a,0x09,0x75,0x63,0x68,0x61,0x72,0x20,0x65,0x74,
0x68,0x65,0x72,0x30,0x6d,0x61,0x63,0x5b,0x5d,0x20,0x3d,0x20,0x7b,0x20,0x32,0x2c,
0x20,0x30,0x2c,0x20,0x30,0x2c,0x20,0x30,0x2c,0x20,0x30,0x2c,0x20,0x31,0x20,0x7d,
0x3b,0x0a,0x0a,0x09,0x2f,0x2a,0x20,0x74,0x68,0x65,0x20,0x65,0x6d,0x75,0x6c,0x61,
0x74,0x65,0x64,0x20,0x70,0x6c,0x69,0x63,0x20,0x64,0x6f,0x65,0x73,0x6e,0x27,0x74,
0x20,0x73,0x65,0x65,0x6d,0x20,0x74,0x6f,0x20,0x66,0x6f,0x6c,0x6c,0x6f,0x77,0x20,
0x74,0x68,0x65,0x20,0x73,0x70,0x65,0x63,0x3b,0x20,0x77,0x65,0x20,0x69,0x67,0x6e,
0x6f,0x72,0x65,0x20,0x69,0x74,0x2e,0x20,0x2a,0x2f,0x0a,0x09,0x53,0x6f,0x63,0x20,
0x73,0x6f,0x63,0x20,0x3d,0x20,0x7b,0x0a,0x09,0x09,0x2e,0x63,0x6c,0x69,0x6e,0x74,
0x09,0x3d,0x20,0x28,0x63,0x68,0x61,0x72,0x20,0x2a,0x29,0x50,0x41,0x43,0x6c,0x69,
0x6e,0x74,0x2c,0x0a,0x09,0x09,0x2e,0x75,0x61,0x72,0x74,0x09,0x3d,0x20,0x28,0x63,
0x68,0x61,0x72,0x20,0x2a,0x29,0x50,0x41,0x55,0x61,0x72,0x74,0x30,0x2c,0x0a,0x09,
0x09,0x2e,0x70,0x6c,0x69,0x63,0x09,0x3d,0x20,0x28,0x63,0x68,0x61,0x72,0x20,0x2a,
0x29,0x30,0x78,0x34,0x30,0x31,0x30,0x30,0x30,0x30,0x30,0x2c,0x0a,0x09,0x09,0x2e,
0x65,0x74,0x68,0x65,0x72,0x5b,0x30,0x5d,0x20,0x3d,0x20,0x28,0x63,0x68,0x61,0x72,
0x20,0x2a,0x29,0x30,0x78,0x34,0x30,0x30,0x31,0x31,0x30,0x30,0x30,0x2c,0x0a,0x09,
0x09,0x2e,0x68,0x6f,0x62,0x62,0x6c,0x65,0x64,0x20,0x3d,0x20,0x30,0x2c,0x09,0x09,
0x2f,0x2a,0x20,0x6f,0x6e,0x6c,0x79,0x20,0x31,0x20,0x68,0x61,0x72,0x74,0x20,0x2a,
0x2f,0x0a,0x09,0x7d,0x3b,0x0a,0x09,0x49,0x6f,0x63,0x6f,0x6e,0x66,0x20,0x73,0x6f,
0x63,0x63,0x6f,0x6e,0x66,0x5b,0x5d,0x20,0x3d,0x20,0x7b,0x20,0x2f,0x2a,0x20,0x64,
0x65,0x76,0x69,0x63,0x65,0x73,0x20,0x77,0x69,0x74,0x68,0x6f,0x75,0x74,0x20,0x64,
0x72,0x69,0x76,0x65,0x72,0x73,0x20,0x74,0x68,0x61,0x74,0x20,0x76,0x6d,0x61,0x70,
0x20,0x74,0x68,0x65,0x69,0x72,0x20,0x72,0x65,0x67,0x73,0x20,0x2a,0x2f,0x0a,0x09,
0x09,0x7b,0x20,0x22,0x63,0x6c,0x69,0x6e,0x74,0x22,0x2c,0x20,0x36,0x34,0x2a,0x4b,
0x42,0x2c,0x20,0x26,0x73,0x6f,0x63,0x2e,0x63,0x6c,0x69,0x6e,0x74,0x2c,0x20,0x7d,
0x2c,0x0a,0x09,0x09,0x7b,0x20,0x22,0x75,0x61,0x72,0x74,0x22,0x2c,0x20,0x20,0x50,
0x47,0x53,0x5a,0x2c,0x20,0x26,0x73,0x6f,0x63,0x2e,0x75,0x61,0x72,0x74,0x2c,0x20,
0x31,0x2c,0x20,0x7d,0x2c,0x0a,0x09,0x09,0x7b,0x20,0x22,0x70,0x6c,0x69,0x63,0x22,
0x2c,0x20,0x20,0x34,0x2a,0x4d,0x42,0x2c,0x20,0x26,0x73,0x6f,0x63,0x2e,0x70,0x6c,
0x69,0x63,0x2c,0x20,0x7d,0x2c,0x09,0x09,0x2f,0x2a,0x20,0x63,0x6f,0x6d,0x6d,0x6f,
0x6e,0x20,0x62,0x75,0x74,0x20,0x73,0x6d,0x61,0x6c,0x6c,0x65,0x72,0x20,0x2a,0x2f,
0x0a,0x09,0x09,0x30,0x0a,0x09,0x7d,0x3b,0x0a,0x09,0x49,0x6f,0x63,0x6f,0x6e,0x66,
0x20,0x69,0x6f,0x63,0x6f,0x6e,0x66,0x73,0x5b,0x5d,0x20,0x3d,0x20,0x7b,0x09,0x2f,
0x2a,0x20,0x64,0x65,0x76,0x69,0x63,0x65,0x73,0x20,0x77,0x68,0x6f,0x73,0x65,0x20,
0x64,0x72,0x69,0x76,0x65,0x72,0x73,0x20,0x76,0x6d,0x61,0x70,0x20,0x74,0x68,0x65,
0x69,0x72,0x20,0x72,0x65,0x67,0x73,0x20,0x2a,0x2f,0x0a,0x09,0x09,0x7b,0x20,0x22,
0x65,0x74,0x68,0x65,0x72,0x22,0x2c,0x20,0x32,0x2a,0x50,0x47,0x53,0x5a,0x2c,0x20,
0x26,0x73,0x6f,0x63,0x2e,0x65,0x74,0x68,0x65,0x72,0x5b,0x30,0x5d,0x2c,0x20,0x32,
0x2c,0x20,0x7d,0x2c,0x0a,0x09,0x09,0x30,0x0a,0x09,0x7d,0x3b,0x0a,0x0a,0x64,0x62,
0x67,0x66,0x6c,0x67,0x0a,0x23,0x09,0x73,0x79,0x73,0x70,0x72,0x6f,0x63,0x09,0x09,
0x27,0x45,0x27,0x0a,0x23,0x09,0x6d,0x70,0x09,0x09,0x27,0x4d,0x27,0x0a,0x23,0x09,
0x70,0x63,0x69,0x09,0x09,0x27,0x50,0x27,0x0a,0x23,0x09,0x61,0x72,0x63,0x68,0x09,
0x09,0x27,0x56,0x27,0x0a,0x23,0x09,0x61,0x72,0x63,0x68,0x72,0x76,0x09,0x09,0x27,
0x56,0x27,0x0a,0x23,0x09,0x61,0x73,0x6d,0x09,0x09,0x27,0x61,0x27,0x0a,0x23,0x09,
0x63,0x68,0x61,0x6e,0x09,0x09,0x27,0x63,0x27,0x0a,0x23,0x09,0x6d,0x6d,0x75,0x09,
0x09,0x27,0x6d,0x27,0x0a,0x23,0x09,0x6d,0x61,0x69,0x6e,0x09,0x09,0x27,0x78,0x27,
0x0a,0x72,0x69,0x73,0x63,0x76,0x36,0x34,0x0a,0x09,0x73,0x74,0x61,0x72,0x74,0x0a,
0x09,0x6c,0x6f,0x77,0x0a,0x09,0x6d,0x74,0x72,0x61,0x70,0x0a,0x09,0x73,0x74,0x72,
0x61,0x70,0x0a,0x09,0x6d,0x63,0x68,0x0a,0x09,0x73,0x61,0x74,0x70,0x0a,0x09,0x6d,
0x61,0x69,0x6e,0x0a,0x09,0x61,0x72,0x63,0x68,0x0a,0x09,0x61,0x72,0x63,0x68,0x72,
0x76,0x0a,0x09,0x63,0x6f,0x6c,0x6f,0x72,0x0a,0x09,0x63,0x73,0x72,0x0a,0x09,0x66,
0x70,0x0a,0x09,0x66,0x70,0x75,0x0a,0x09,0x69,0x6f,0x63,0x6f,0x6e,0x66,0x0a,0x23,
0x09,0x6b,0x62,0x64,0x0a,0x09,0x6d,0x61,0x70,0x0a,0x09,0x6d,0x65,0x6d,0x6f,0x72,
0x79,0x0a,0x09,0x6d,0x6d,0x75,0x0a,0x09,0x66,0x61,0x6b,0x65,0x70,0x63,0x69,0x0a,
0x09,0x73,0x61,0x6e,0x65,0x0a,0x09,0x66,0x61,0x6b,0x65,0x73,0x62,0x69,0x0a,0x09,
0x73,0x62,0x69,0x0a,0x09,0x74,0x72,0x61,0x70,0x0a,0x70,0x6f,0x72,0x74,0x0a,0x09,
0x61,0x6c,0x61,0x72,0x6d,0x0a,0x09,0x61,0x6c,0x6c,0x6f,0x63,0x62,0x0a,0x09,0x61,
0x77,0x61,0x69,0x74,0x62,0x69,0x74,0x0a,0x09,0x62,0x79,0x74,0x65,0x73,0x65,0x72,
0x69,0x61,0x6c,0x0a,0x09,0x63,0x68,0x61,0x6e,0x0a,0x09,0x63,0x74,0x79,0x70,0x65,
0x0a,0x09,0x64,0x65,0x76,0x0a,0x09,0x64,0x65,0x76,0x74,0x61,0x62,0x0a,0x09,0x65,
0x64,0x66,0x0a,0x09,0x66,0x61,0x75,0x6c,0x74,0x0a,0x09,0x69,0x6d,0x61,0x67,0x65,
0x0a,0x23,0x09,0x6c,0x61,0x74,0x69,0x6e,0x31,0x0a,0x09,0x6c,0x6f,0x67,0x70,0x6f,
0x77,0x0a,0x09,0x70,0x61,0x67,0x65,0x0a,0x09,0x70,0x61,0x72,0x73,0x65,0x0a,0x09,
0x70,0x67,0x72,0x70,0x0a,0x09,0x70,0x6f,0x72,0x74,0x63,0x6c,0x6f,0x63,0x6b,0x0a,
0x09,0x70,0x72,0x66,0x0a,0x09,0x70,0x72,0x69,0x6e,0x74,0x0a,0x09,0x70,0x72,0x6f,
0x63,0x0a,0x09,0x70,0x73,0x0a,0x09,0x71,0x69,0x6f,0x0a,0x09,0x71,0x6c,0x6f,0x63,
0x6b,0x0a,0x09,0x71,0x6d,0x61,0x6c,0x6c,0x6f,0x63,0x0a,0x09,0x72,0x61,0x6e,0x64,
0x6f,0x6d,0x0a,0x09,0x72,0x65,0x62,0x6f,0x6f,0x74,0x63,0x6d,0x64,0x0a,0x09,0x73,
0x65,0x67,0x6d,0x65,0x6e,0x74,0x0a,0x09,0x73,0x79,0x73,0x61,0x75,0x74,0x68,0x0a,
0x09,0x73,0x79,0x73,0x63,0x61,0x6c,0x6c,0x0a,0x09,0x73,0x79,0x73,0x63,0x61,0x6c,
0x6c,0x66,0x6d,0x74,0x0a,0x09,0x73,0x79,0x73,0x66,0x69,0x6c,0x65,0x0a,0x09,0x73,
0x79,0x73,0x70,0x72,0x6f,0x63,0x0a,0x09,0x73,0x79,0x73,0x73,0x65,0x67,0x0a,0x09,
0x73,0x79,0x73,0x74,0x61,0x62,0x0a,0x09,0x74,0x61,0x73,0x6c,0x6f,0x63,0x6b,0x0a,
0x09,0x74,0x6f,0x64,0x0a,0x09,0x77,0x61,0x74,0x65,0x72,0x6d,0x61,0x72,0x6b,0x73,
0x0a,0x0a,0x23,0x0a,0x23,0x20,0x63,0x6f,0x6e,0x74,0x72,0x6f,0x6c,0x73,0x20,0x77,
0x68,0x69,0x63,0x68,0x20,0x64,0x69,0x72,0x65,0x63,0x74,0x6f,0x72,0x69,0x65,0x73,
0x20,0x74,0x6f,0x20,0x6c,0x6f,0x6f,0x6b,0x20,0x69,0x6e,0x20,0x66,0x6f,0x72,0x20,
0x73,0x6f,0x75,0x72,0x63,0x65,0x20,0x66,0x69,0x6c,0x65,0x73,0x3b,0x20,0x6d,0x75,
0x73,0x74,0x20,0x62,0x65,0x20,0x72,0x65,0x6c,0x61,0x74,0x69,0x76,0x65,0x20,0x74,
0x6f,0x20,0x2e,0x2e,0x0a,0x23,0x0a,0x64,0x69,0x72,0x0a,0x09,0x69,0x70,0x0a,0x09,
0x70,0x6f,0x72,0x74,0x0a,

0,
};
