/*
 * mapping twixt kernel-virtual and physical addresses.
 * trivial mapping on rv32.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

void*
KADDR(uintptr pa)
{
	if (pa >= (uintptr)KZERO)
		return (void *)(pa - physmem + KZERO);
	panic("KADDR: pa %#p not in kernel from %#p", pa, getcallerpc(&pa));
	notreached();
}

uintmem
PADDR(void* va)
{
	if((uintptr)va >= KSEG0 && (uintptr)va < KSEG0 + physsize)
		return (uintptr)va - KZERO + physmem;
	/* probably in user space */
	panic("PADDR: va %#p from %#p", va, getcallerpc(&va));
	notreached();
}

/*
 * KMap is used to map individual pages into virtual memory.
 * It is rare to have more than a few KMaps at a time (in the
 * absence of interrupts, only two at a time are ever used,
 * but interrupts can stack).  The mappings are local to a process,
 * so we can use the same range of virtual address space for
 * all processes without any coordination.
 */
/* return a working virtual address for page->pa */
KMap*
kmap(Page* page)
{
	void *va;

	if (page == nil)
		panic("kmap: nil page");
	if (page->pa == 0)
		panic("kmap: nil page->pa");
	va = KADDR(page->pa);
//	print("kmap(%#llux) @ %#p: %#p %#p\n",
//		page->pa, getcallerpc(&page), page->pa, va);
	if (mmuphysaddr((uintptr)va) == ~0ull)
		print("kmap: no mapped pa for va %#p (from pa %#p)\n",
			va, page->pa);
	return va;
}

/*
 * Return the number of bytes that can be accessed via KADDR(pa).
 * If pa is not a valid argument to KADDR, return 0.
 */
uintptr
cankaddr(uintptr pa)
{
	if(pa >= (uintptr)KZERO && pa < KZERO + physsize)
		return physmem + physsize - pa;
	return 0;
}
