/*
 * Reboot trampoline code.  Copies new kernel into place over old one.
 * This is copied to sys->reboottramp before executing it.
 */
#include "mem.h"
#include "riscvl.h"
#include "start.h"

#define SUPERENTRY 0x10	/* offset from _main of super entry in start.s */

/* dedicated registers for reboot */
#define ENTRY	17
#define SEG0MASK 18
#define RMASK	19

/*
 * Turn off MMU, then memmove the new kernel to its correct location
 * in physical addresses.  Then jump to the start of the kernel.
 * we are called in the identity map with
 *
 *	(*tramp)(Reboot *);
 *
 * We must be entered in machine mode in order to start the next kernel
 * in machine mode.  An environment call from super mode will get us there.
 */
TEXT main(SB), $0
#ifdef MACHINE_MODE
	CSRRC	CSR(MSTATUS), $Mie, R0		/* splhi */
#endif
	CSRRC	CSR(SSTATUS), $(Mie|Sie), R0	/* splhi */
	MOV	$setSB(SB), R3
	MOV	$PAUart0, R(UART0)
CONSPUT($'=')
	MOV	$((1ull<<32)-1), R(RMASK) /* 32-bit mask */

	MOV	R(ARG), R(SYSRB)	/* physical &sys->Reboot */

	MOV	R0, R9
	LOADSATP(R9)			/* mmu off in super */
//	MOV	$setSB(SB), R3

	MOV	R0, newkern(SB)

	MOV	$strap(SB), R(TMP)
	MOV	R(TMP), CSR(STVEC)
	MOV	R(TMP), CSR(SEPC)

	MOVW	MMACHNO(R(MACH)), R(MACHNO)

	/* copy args out of KSEG0 space */
	/* entry is only 32 bits */
	MOV	RBENTRY(R(SYSRB)), R9
	AND	R(RMASK), R9		/* destination va; make it physical */
	MOV	R9, R(ENTRY)		/* also entry point */
	MOVBU	MMACHMODE(R(MACH)), R(MACHMODE)
	BNE	R(MACHMODE), machentry
	ADD	$SUPERENTRY, R(ENTRY)	/* start at super entry point */
machentry:
CONSPUT($'3')
	MOV	RBCODE(R(SYSRB)), R10	/* source va */
	MOV	RBSIZE(R(SYSRB)), R11	/* byte count */

	/* copy before possibly overwriting Mach */
	MOVWU	MACHHARTID(R(MACH)), R(HARTID)	/* emulate SBI start-up */

	BNE	R(MACHNO), secondary

/*
 * the source and destination may overlap.
 * determine whether to copy forward or backwards
 */
	MOV	$BY2WD, R13
	BGTU	R9, R10, forw		/* src > dest, copy forward */
	ADD	R10, R11, R14
	BLEU	R9, R14, forw		/* src+len <= dest, copy forward */
	/* copy new kernel into place */
	ADD	R11, R9			/* start at the ends */
	ADD	R11, R10
	SUB	R13, R9
	SUB	R13, R10
back:
	MOV	(R10), R14
	MOV	R14, (R9)
	SUB	R13, R10
	SUB	R13, R9
	SUB	R13, R11
	BGT	R11, back
	JMP	done
forw:
	MOV	(R10), R14
	MOV	R14, (R9)
	ADD	R13, R10
	ADD	R13, R9
	SUB	R13, R11
	BGT	R11, forw
done:
	FENCE
	FENCE_I
CONSPUT($'&')
	MOV	$RBFLAGSET, R(TMP2)
	MOV	R(TMP2), newkern(SB)		/* release other cpus */
	JMP	newkernel

/*
 * secondary cpus spin here until sys->newkernel is true,
 * indicating that the new kernel is in place.
 */
secondary:
CONSPUT($'W')
	/* wait for cpu0 to load the new kernel */
	MOV	$newkern(SB), R(TMP2)
	MOV	$RBFLAGSET, R10
awaitkern:
	FENCE
	MOV	(R(TMP2)), R(TMP)
	BNE	R(TMP), R10, awaitkern

/*
 * JMP to kernel entry point.  Note the true kernel entry point is
 * the virtual address KZERO|R(ENTRY), but on some systems, this must wait until
 * the MMU is enabled by the kernel in start.s.
 */
newkernel:
	FENCE
	FENCE_I
CONSPUT($'|')
	MOV	R(HARTID), R10			/* emulate SBI start up */
#ifdef MACHINE_MODE
	MOV	R(ENTRY), CSR(MTVEC)
	MOV	R(ENTRY), CSR(MEPC)
#endif
	MOV	R(ENTRY), CSR(STVEC)
	MOV	R(ENTRY), CSR(SEPC)
/**/	JMP	(R(ENTRY))

TEXT strap(SB), 1, $-4
	CONSPUT($'!')
	MOV	CSR(SCAUSE), R8
	ADD	$'0', R8
	CONSPUT(R8)
wfiloop:
	FENCE
	WFI
	JMP	wfiloop

	GLOBL	newkern(SB), $8
	DATA	newkern(SB)/8, $0
