/*
 * risc-v definitions
 */
#define SYSTEM		0x73			/* op code */
#define ECALLINST	SYSTEM
#define ISCOMPRESSED(inst) (((inst) & MASK(2)) != 3)

/* registers */
#define LINK	R1
#define USER	6			/* up-> */
#define MACH	7			/* m-> */
#define ARG	8

#define UREGSIZE (XLEN*(32+6))	/* room for pc, regs, csrs, mode, padding */

/*
 * byte offsets into Mach, m, (mainly for m->regsave), for assembler,
 * where XLEN is defined.
 */
#define MMACHNO		0
#define SPLPC		XLEN
#define MACHPROC	(2*XLEN)
#define SAVESR2		(3*XLEN)
#define SAVESR4		(4*XLEN)
#define SAVESR6		(5*XLEN)
#define SAVEMR2		(6*XLEN)
#define SAVEMR4		(7*XLEN)
#define SAVEMR6		(8*XLEN)
#define ONLINE		(9*XLEN)
#define MACHHARTID	(10*XLEN)
#define PLICCXTX	(11*XLEN)
#define MMACHMODE	(12*XLEN)

/*
 * byte offsets into Sys->Reboot.
 */
#define RBTRAMP	0
#define RBENTRY XLEN
#define RBCODE	(2*XLEN)
#define RBSIZE	(3*XLEN)
#define RBSYS	(4*XLEN)
#define RBSTALL	(5*XLEN)
#define RBSATP	(6*XLEN)

/* magic numbers for flags */
#define RBFLAGSET (0xecce|1)		/* ensure low bit set for begin.c */

enum {					/* CSRs */
	FFLAGS	= 1,
	FRM	= 2,
	FCSR	= 3,

	SUPERBASE = 0x100,
	SSTATUS	= 0x100,
	SEDELEG	= 0x102,
	SIDELEG	= 0x103,
	SIE	= 0x104,
	STVEC	= 0x105,
	SCOUNTEREN = 0x306,
	SSCRATCH = 0x140,
	SEPC	= 0x141,
	SCAUSE	= 0x142,
	STVAL	= 0x143,
	SIP	= 0x144,
	SATP	= 0x180,		/* page table root, paging mode */

	MCHBASE	= 0x300,
	MSTATUS	= 0x300,
	MISA	= 0x301,
	MEDELEG	= 0x302,
	MIDELEG	= 0x303,
	MIE	= 0x304,
	MTVEC	= 0x305,
	MCOUNTEREN = 0x306,
	MSTATUSH = 0x310,		/* RV32 only */
	MSCRATCH = 0x340,
	MEPC	= 0x341,
	MCAUSE	= 0x342,
	MTVAL	= 0x343,
	MIP	= 0x344,

	/* cycle counters, timers, clocks, etc. */
	MCYCLELO = 0xb00,		/* rw in mach mode */
	MCYCLEHI = 0xb80,		/* RV32 only */
	CYCLO	= 0xc00,
//	TIMELO	= 0xc01,		/* not implemented */
	CYCHI	= 0xc80,		/* RV32 only */
//	TIMEHI	= 0xc81,		/* RV32 only */

	MVENDORID = 0xf11,
	MARCHID	= 0xf12,
	MIMPLID	= 0xf13,
	MHARTID	= 0xf14,
};

enum {				/* mstatus, status bits for RV32 */
	Sie	= (1<<1),	/* supervisor-mode intr enable */
	Mie	= (1<<3),	/* machine-mode intr enable */
	Spie	= (1<<5),	/* previous Sie */
	Ube	= (1<<6),	/* user big endian */
	Mpie	= (1<<7),	/* previous Mie (machine only) */
	Spp	= (1<<8),	/* previous priv. mode (super) */
	Mpp	= (3<<11),	/* previous privilege mode (machine only) */
	Mppuser	= (0<<11),
	Mppsuper = (1<<11),
	Mpphype	= (2<<11),
	Mppmach	= (3<<11),
	Fsshft	= 13,
	Fs	= (3<<Fsshft),
	Xsshft	= 15,
	Xs	= (3<<Xsshft),
	Mprv	= (1<<17),	/* use Mpp priv.s for memory access (machine) */
	Sum	= (1<<18),	/* super user memory access (super) */
	Mxr	= (1<<19),	/* make executable readable (super) */
	Tvm	= (1<<20),	/* trap vm instrs. (machine) */
	Tw	= (1<<21),	/* timeout wfi (machine) */
	Tsr	= (1<<22),	/* trap SRET in super mode */
	Sd	= (1<<31),	/* dirty state */

	/* mstatush (rv32 only) */
	Sbe	= (1<<4),	/* super big endian */
	Mbe	= (1<<5),	/* machine big endian */
};

enum {				/* xs & fs values in ?status */
	Off	= 0,
	Initial	= 1,
	Clean	= 2,
	Dirty	= 3,
	Fsmask	= 3,
};

enum {			/* [ms]i[ep] bits for RV32.  only S* bits in si[ep] */
	Ssie	= (1<<1),	/* super sw intr enable */
	Msie	= (1<<3),	/* machine sw intr eanble */
	Stie	= (1<<5),	/* super timer intr enable */
	Mtie	= (1<<7),	/* machine timer intr enable */
	Seie	= (1<<9),	/* super ext intr enable (sie); global (mie) */
	Meie	= (1<<11),	/* machine global interrupt enable */
	Lie0	= (1<<16),	/* local intr 0 enable */
};

enum {
	Mxlen32	= 1,
	Mxlen64 = 2,
	Mxlenmask = 3,
};

enum {
	Rv32intr	= (1<<31), /* interrupt, not exception */
	Rv64intr	= (1ull<<63), /* interrupt, not exception */
};

enum {				/* cause exception codes for interrupts */
	Supswintr	= 1,
	Mchswintr	= 3,
	Suptmrintr	= 5,
	Mchtmrintr	= 7,
	Supextintr	= 9,
	Mchextintr	= 11,
	Nlintr		= 32,	/* # of local interrupts */
};

enum {				/* cause exception codes for traps */
	Instaddralign	= 0,
	Instaccess	= 1,
	Illinst		= 2,
	Breakpt		= 3,
	Loadaddralign	= 4,
	Loadaccess	= 5,
	Storeaddralign	= 6,
	Storeaccess	= 7,
	Envcalluser	= 8,		/* system call from user mode */
	Envcallsup	= 9,		/* environment call from super mode */
	Reserved10	= 10,
	Envcallmch	= 11,		/* environment call from machine mode */
	Instpage	= 12,		/* page faults */
	Loadpage	= 13,
	Reserved14	= 14,
	Storepage	= 15,
	Nexc		= 16,
};

enum {					/* CSR(SATP) */
	Pagingon	= (1<<31),	/* Sv32 paging enable */
};

enum {					/* PTE; see mem.h PTE* */
	PteP		= 0x01,		/* Present/valid */
	PteR		= 0x02,		/* Read */
	PteW		= 0x04,		/* Write */
	PteX		= 0x08,		/* Execute */
	PteU		= 0x10,	/* accessible to U mode; only usable on leaves */
	PteG		= 0x20,		/* Global */
	PteA		= 0x40,		/* Accessed */
	PteD		= 0x80,		/* Dirty */
	PteRSW		= (MASK(3)<<8),	/* reserved for sw use */
	PtePS		= (1<<8),	/* big page, not 4K */
	PtePtr		= (1<<9),	/* ptr to next llvl pt (PteLeaf=0) */

	PteLeaf		= 0xe,		/* rwx; if 0, next level; non-0 leaf */
	/* some implementations require these in leaf ptes */
	Pteleafmand	= (PteA|PteD),
	PteAttrs	= (MASK(6)|PteRSW),
};

enum {
	Vaddrbits	= 32,		/* for RV32 */
};

/*
 * plic interrupt sources
 *
 * intrs/irqs (via plic)
 * name		global int	irq
 * mac0		51		64
 * mac1		57		70
 * rtc wakeup	67		80
 * rtc match	68		81
 * timer1	69		82
 * timer2	70		83
 * usb dma	73		86
 * usb mc	74		87
 * mmuart0	77		90
 */
enum Plicirqs {					/* Plicirqs */
	/*
	 * there are 1024 possible interrupt sources, but we don't
	 * use them all.
	 */
	Ngintr		= 187,

	Gliirqdiff	= 13,	/* 1 zero + 4 l2 cache + 8 dma */
};

/*
 * SBI crud
 */
#define HSM	0x48534dll		/* v0.2 extension */
