#define SYSTEM		0x73			/* op code */
#define ECALLINST	SYSTEM
#define ISCOMPRESSED(inst) (((inst) & MASK(2)) != 3)

#define LINK	R1
#define USER	6			/* up-> */
#define MACH	7			/* m-> */
#define ARG	8

#define UREGSIZE (XLEN*(32+6))	/* room for pc, regs, csrs, mode, padding */

#define MMACHNO		0
#define SPLPC		XLEN
#define MACHPROC	(2*XLEN)
#define SAVESR2		(3*XLEN)
#define SAVESR4		(4*XLEN)
#define SAVESR6		(5*XLEN)
#define SAVEMR2		(6*XLEN)
#define SAVEMR4		(7*XLEN)
#define SAVEMR6		(8*XLEN)
#define ONLINE		(9*XLEN)
#define MACHHARTID	(10*XLEN)
#define PLICCXTX	(11*XLEN)
#define MMACHMODE	(12*XLEN)

#define RBTRAMP	0
#define RBENTRY XLEN
#define RBCODE	(2*XLEN)
#define RBSIZE	(3*XLEN)
#define RBSYS	(4*XLEN)
#define RBSTALL	(5*XLEN)
#define RBSATP	(6*XLEN)

#define RBFLAGSET (0xecce|1)		/* ensure low bit set for begin.c */

#define FFLAGS		1
#define FRM		2
#define FCSR		3

#define SUPERBASE	0x100
#define SSTATUS		0x100
#define SEDELEG		0x102
#define SIDELEG		0x103
#define SIE		0x104
#define STVEC		0x105
#define SCOUNTEREN	0x306
#define SSCRATCH	0x140
#define SEPC		0x141
#define SCAUSE		0x142
#define STVAL		0x143
#define SIP		0x144
#define SATP		0x180			/* page table root, paging mode */

#define MCHBASE		0x300
#define MSTATUS		0x300
#define MISA		0x301
#define MEDELEG		0x302
#define MIDELEG		0x303
#define MIE		0x304
#define MTVEC		0x305
#define MCOUNTEREN	0x306
#define MSTATUSH	0x310			/* RV32 only */
#define MSCRATCH	0x340
#define MEPC		0x341
#define MCAUSE		0x342
#define MTVAL		0x343
#define MIP		0x344

#define MCYCLELO	0xb00			/* rw in mach mode */
#define MCYCLEHI	0xb80			/* RV32 only */
#define CYCLO		0xc00
#define CYCHI		0xc80			/* RV32 only */

#define MVENDORID	0xf11
#define MARCHID		0xf12
#define MIMPLID		0xf13
#define MHARTID		0xf14

#define Sie		(1<<1)			/* supervisor-mode intr enable */
#define Mie		(1<<3)			/* machine-mode intr enable */
#define Spie		(1<<5)			/* previous Sie */
#define Ube		(1<<6)			/* user big endian */
#define Mpie		(1<<7)			/* previous Mie (machine only) */
#define Spp		(1<<8)			/* previous priv. mode (super) */
#define Mpp		(3<<11)			/* previous privilege mode (machine only) */
#define Mppuser		(0<<11)
#define Mppsuper	(1<<11)
#define Mpphype		(2<<11)
#define Mppmach		(3<<11)
#define Fsshft		13
#define Fs		(3<<Fsshft)
#define Xsshft		15
#define Xs		(3<<Xsshft)
#define Mprv		(1<<17)			/* use Mpp priv.s for memory access (machine) */
#define Sum		(1<<18)			/* super user memory access (super) */
#define Mxr		(1<<19)			/* make executable readable (super) */
#define Tvm		(1<<20)			/* trap vm instrs. (machine) */
#define Tw		(1<<21)			/* timeout wfi (machine) */
#define Tsr		(1<<22)			/* trap SRET in super mode */
#define Sd		(1<<31)			/* dirty state */

#define Sbe		(1<<4)			/* super big endian */
#define Mbe		(1<<5)			/* machine big endian */

#define Off		0
#define Initial		1
#define Clean		2
#define Dirty		3
#define Fsmask		3

#define Ssie		(1<<1)			/* super sw intr enable */
#define Msie		(1<<3)			/* machine sw intr eanble */
#define Stie		(1<<5)			/* super timer intr enable */
#define Mtie		(1<<7)			/* machine timer intr enable */
#define Seie		(1<<9)			/* super ext intr enable (sie); global (mie) */
#define Meie		(1<<11)			/* machine global interrupt enable */
#define Lie0		(1<<16)			/* local intr 0 enable */

#define Mxlen32		1
#define Mxlen64		2
#define Mxlenmask	3

#define Rv32intr	(1<<31)			/* interrupt, not exception */
#define Rv64intr	(1ull<<63)		/* interrupt, not exception */

#define Supswintr	1
#define Mchswintr	3
#define Suptmrintr	5
#define Mchtmrintr	7
#define Supextintr	9
#define Mchextintr	11
#define Nlintr		32			/* # of local interrupts */

#define Instaddralign	0
#define Instaccess	1
#define Illinst		2
#define Breakpt		3
#define Loadaddralign	4
#define Loadaccess	5
#define Storeaddralign	6
#define Storeaccess	7
#define Envcalluser	8			/* system call from user mode */
#define Envcallsup	9			/* environment call from super mode */
#define Reserved10	10
#define Envcallmch	11			/* environment call from machine mode */
#define Instpage	12			/* page faults */
#define Loadpage	13
#define Reserved14	14
#define Storepage	15
#define Nexc		16

#define Pagingon	(1<<31)			/* Sv32 paging enable */

#define PteP		0x01			/* Present/valid */
#define PteR		0x02			/* Read */
#define PteW		0x04			/* Write */
#define PteX		0x08			/* Execute */
#define PteU		0x10			/* accessible to U mode; only usable on leaves */
#define PteG		0x20			/* Global */
#define PteA		0x40			/* Accessed */
#define PteD		0x80			/* Dirty */
#define PteRSW		(MASK(3)<<8)		/* reserved for sw use */
#define PtePS		(1<<8)			/* big page, not 4K */
#define PtePtr		(1<<9)			/* ptr to next llvl pt (PteLeaf=0) */

#define PteLeaf		0xe			/* rwx; if 0, next level; non-0 leaf */
#define Pteleafmand	(PteA|PteD)
#define PteAttrs	(MASK(6)|PteRSW)

#define Vaddrbits	32			/* for RV32 */



#define HSM	0x48534dll		/* v0.2 extension */
#define IS_ASM

/*
 * Macros for accessing page table entries; change the
 * C-style array-index macros into a page table byte offset.
 * RV32 PTEs are longs, so shift by 2 bits.
 */
// #define PTROOTO(v)	((PTLX((v), 3))<<2)
// #define PDPO(v)	((PTLX((v), 2))<<2)
#define PDO(v)		((PTLX((v), 1))<<2)
#define PTO(v)		((PTLX((v), 0))<<2)

#define SAVECONSREGS \
	MOV	$setSB(SB), R3; \
	MOV	R20, dbsave(SB); \
	MOV	R21, dbsave+XLEN(SB); \
	MOV	R22, dbsave+(2*XLEN)(SB)

#define RESTCONSREGS \
	MOV	$setSB(SB), R3; \
	MOV	dbsave(SB), R20; \
	MOV	dbsave+XLEN(SB), R21; \
	MOV	dbsave+(2*XLEN)(SB), R22

	GLOBL	dbsave(SB), $(3*XLEN)

#ifndef MASK
#define MASK(w)		((1u  <<(w)) - 1)
#define VMASK(w)	((1ull<<(w)) - 1)
#endif

/* instructions not yet known to ia */
#define ECALL	SYS	$0
#define EBREAK	SYS	$1
#define URET	SYS	$2
#define SRET	SYS	$0x102
#define WFI	SYS	$0x105
#define MRET	SYS	$0x302

/* atomics */
#define	Amoadd	0
#define	Amoswap	1
#define	Amolr	2
#define	Amosc	3
#define	Amoxor	4
#define	Amoor	010
#define	Amoand	014
#define	Amomin	020
#define	Amomax	024
#define	Amominu	030
#define	Amomaxu	034

/* AMO operand widths */
#define	Amow	2		/* long */
#define	Amod	3		/* vlong */

#define FENCE	WORD $(0xf | MASK(8)<<20)  /* all i/o, mem ops before & after */
#define AQ	(1<<26)			/* acquire */
#define RL	(1<<25)			/* release */

/* atomically (rd = (rs1); (rs1) op= rs2) */
#define AMOW(func, opts, rs2, rs1, rd) \
	WORD $(((func)<<27)|((rs2)<<20)|((rs1)<<15)|(Amow<<12)|((rd)<<7)|057|opts)
#define LRW(rs1, rd)		AMOW(Amolr, AQ, 0, rs1, rd)
#define SCW(rs2, rs1, rd)	AMOW(Amosc, AQ|RL, rs2, rs1, rd)

#define SYNC	FENCE
#define FENCE_I	WORD	$(1<<12|0xf)
#define SFENCE_VMA(rs2, rs1) WORD $(011<<25|(rs2)<<20|(rs1)<<15|0<<7|SYSTEM)

/* rv64 instructions */
#define AMOD(func, rs2, rs1, rd) \
	WORD $(((func)<<27)|((rs2)<<20)|((rs1)<<15)|(Amod<<12)|((rd)<<7)|057)
#define	SLLI64(rs2, rs1, rd) \
	WORD $((rs2)<<20 | (rs1)<<15 | 1<<12 | (rd)<<7 | 0x13)
#define	SRLI64(rs2, rs1, rd) \
	WORD $((rs2)<<20 | (rs1)<<15 | 5<<12 | (rd)<<7 | 0x13)

/*
 * strap and mtrap macros
 */
#define PUSH(n)		MOV R(n), ((n)*XLEN)(R2)
#define POP(n)		MOV ((n)*XLEN)(R2), R(n)

#define SAVER1_31 \
	PUSH(1); PUSH(2); PUSH(3); PUSH(4); PUSH(5); PUSH(6); PUSH(7); PUSH(8);\
	PUSH(9); PUSH(10); PUSH(11); PUSH(12); PUSH(13); PUSH(14); PUSH(15);\
	PUSH(16); PUSH(17); PUSH(18); PUSH(19); PUSH(20); PUSH(21); PUSH(22);\
	PUSH(23); PUSH(24); PUSH(25); PUSH(26); PUSH(27); PUSH(28); PUSH(29);\
	PUSH(30); PUSH(31)

#define POPR8_31 \
	POP(8); POP(9); POP(10); POP(11); POP(12); POP(13); POP(14); POP(15);\
	POP(16); POP(17); POP(18); POP(19); POP(20); POP(21); POP(22); POP(23);\
	POP(24); POP(25); POP(26); POP(27); POP(28); POP(29); POP(30); POP(31)

#undef	PTLX
#define PTLX(v, l) (((v) >> ((l)*PTSHFT+PGSHFT)) & ((1<<PTSHFT)-1))
