#define IS_ASM

/*
 * Macros for accessing page table entries; change the
 * C-style array-index macros into a page table byte offset.
 * RV32 PTEs are longs, so shift by 2 bits.
 */
// #define PTROOTO(v)	((PTLX((v), 3))<<2)
// #define PDPO(v)	((PTLX((v), 2))<<2)
#define PDO(v)		((PTLX((v), 1))<<2)
#define PTO(v)		((PTLX((v), 0))<<2)

#define SAVECONSREGS \
	MOV	$setSB(SB), R3; \
	MOV	R20, dbsave(SB); \
	MOV	R21, dbsave+XLEN(SB); \
	MOV	R22, dbsave+(2*XLEN)(SB)

#define RESTCONSREGS \
	MOV	$setSB(SB), R3; \
	MOV	dbsave(SB), R20; \
	MOV	dbsave+XLEN(SB), R21; \
	MOV	dbsave+(2*XLEN)(SB), R22

	GLOBL	dbsave(SB), $(3*XLEN)

#ifndef MASK
#define MASK(w)		((1u  <<(w)) - 1)
#define VMASK(w)	((1ull<<(w)) - 1)
#endif

/* instructions not yet known to ia */
#define ECALL	SYS	$0
#define EBREAK	SYS	$1
#define URET	SYS	$2
#define SRET	SYS	$0x102
#define WFI	SYS	$0x105
#define MRET	SYS	$0x302

/* atomics */
#define	Amoadd	0
#define	Amoswap	1
#define	Amolr	2
#define	Amosc	3
#define	Amoxor	4
#define	Amoor	010
#define	Amoand	014
#define	Amomin	020
#define	Amomax	024
#define	Amominu	030
#define	Amomaxu	034

/* AMO operand widths */
#define	Amow	2		/* long */
#define	Amod	3		/* vlong */

#define FENCE	WORD $(0xf | MASK(8)<<20)  /* all i/o, mem ops before & after */
#define AQ	(1<<26)			/* acquire */
#define RL	(1<<25)			/* release */

/* atomically (rd = (rs1); (rs1) op= rs2) */
#define AMOW(func, opts, rs2, rs1, rd) \
	WORD $(((func)<<27)|((rs2)<<20)|((rs1)<<15)|(Amow<<12)|((rd)<<7)|057|opts)
#define LRW(rs1, rd)		AMOW(Amolr, AQ, 0, rs1, rd)
#define SCW(rs2, rs1, rd)	AMOW(Amosc, AQ|RL, rs2, rs1, rd)

#define SYNC	FENCE
#define FENCE_I	WORD	$(1<<12|0xf)
#define SFENCE_VMA(rs2, rs1) WORD $(011<<25|(rs2)<<20|(rs1)<<15|0<<7|SYSTEM)

/* rv64 instructions */
#define AMOD(func, rs2, rs1, rd) \
	WORD $(((func)<<27)|((rs2)<<20)|((rs1)<<15)|(Amod<<12)|((rd)<<7)|057)
#define	SLLI64(rs2, rs1, rd) \
	WORD $((rs2)<<20 | (rs1)<<15 | 1<<12 | (rd)<<7 | 0x13)
#define	SRLI64(rs2, rs1, rd) \
	WORD $((rs2)<<20 | (rs1)<<15 | 5<<12 | (rd)<<7 | 0x13)

/*
 * strap and mtrap macros
 */
#define PUSH(n)		MOV R(n), ((n)*XLEN)(R2)
#define POP(n)		MOV ((n)*XLEN)(R2), R(n)

#define SAVER1_31 \
	PUSH(1); PUSH(2); PUSH(3); PUSH(4); PUSH(5); PUSH(6); PUSH(7); PUSH(8);\
	PUSH(9); PUSH(10); PUSH(11); PUSH(12); PUSH(13); PUSH(14); PUSH(15);\
	PUSH(16); PUSH(17); PUSH(18); PUSH(19); PUSH(20); PUSH(21); PUSH(22);\
	PUSH(23); PUSH(24); PUSH(25); PUSH(26); PUSH(27); PUSH(28); PUSH(29);\
	PUSH(30); PUSH(31)

#define POPR8_31 \
	POP(8); POP(9); POP(10); POP(11); POP(12); POP(13); POP(14); POP(15);\
	POP(16); POP(17); POP(18); POP(19); POP(20); POP(21); POP(22); POP(23);\
	POP(24); POP(25); POP(26); POP(27); POP(28); POP(29); POP(30); POP(31)

#undef	PTLX
#define PTLX(v, l) (((v) >> ((l)*PTSHFT+PGSHFT)) & ((1<<PTSHFT)-1))
