/*
 * verify that various things work as expected:
 * data segment alignment, varargs, long dereferences, clock ticking.
 */

#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

/* verify varargs machinery */
static void
varargck(char *fmt, ...)
{
	uint i, j;
	uchar c, c2;
	uvlong vl, vl2;
	va_list list;

	va_start(list, fmt);
	i = 0;
	if (fmt[0] == '\0')
		i = va_arg(list, uint);
	j = va_arg(list, uint);
	c  = va_arg(list, uchar);
	vl = va_arg(list, uvlong);
	c2  = va_arg(list, uchar);
	vl2 = va_arg(list, uvlong);
	va_end(list);

	if (fmt[0] == '\0')
		if (i != (1<<30))
			iprint("varargck %s int i vararg botch\n", fmt);
	if (j != (2<<28))
		iprint("varargck %s int j vararg botch\n", fmt);
	if (c != 0222)
		iprint("varargck %s char c vararg botch\n", fmt);
	if (vl != (3ll<<40 | 5<<20))
		iprint("varargck %s vlong vl vararg botch\n", fmt);
	if (c2 != 0111)
		iprint("varargck %s char c2 vararg botch\n", fmt);
	if (vl2 != (7ll<<40 | 12<<20))
		iprint("varargck %s vlong vl2 vararg botch\n", fmt);
}

static long
negfunc(void)
{
	long l = -1;

	return l;
}

static void
negint(void)
{
	long l = -1;
	long *lp;

	lp = &l;
	if (*lp >= 0)
		print("*lp >= 0!\n");
	if (negfunc() >= 0)
		print("negfunc() >= 0!\n");
}

static vlong zvl;		/* in bss */

void
sanity(void)
{
	static Lock testlock;
	static ulong align = 123456;

	if (align != 123456)
		panic("mis-aligned data segment; expected %#x saw %#lux",
			123456, align);
	if (zvl != 0)
		panic("mis-initialized bss; expected 0 saw %#llux", zvl);

	assert(sys != 0);
	if (KTZERO != (uintptr)_main)
		panic("KTZERO %#p != _main %#p", KTZERO, _main);
	if ((uintptr)sys < KSEG0)
		panic("sys %#p < KSEG0 %#p", sys, KSEG0);

	lock(&testlock);
	unlock(&testlock);

	varargck("", 1<<30, 2<<28, 0222, 3ll<<40 | 5<<20, 0111,
		(7ll<<40 | 12<<20));
	varargck("2", 2<<28, 0222, 3ll<<40 | 5<<20, 0111,
		(7ll<<40 | 12<<20));
	varargck("3", 2<<28, 0222, 3ll<<40 | 5<<20, 0111,
		(7ll<<40 | 12<<20), 0);
	negint();
}

int
clocksanity(void)
{
	int i;
	ulong ticks;
	uvlong omtime;
	Clint *clint;

	clint = clintaddr();
	clockoff(clint);
	ticks = m->ticks;
	if (m->machno == 0)
		clint->mtime = 0;	/* try to reset; may not take */
	setclinttmr(clint, VMASK(63));
	coherence();
	clrstip();

	omtime = clint->mtime;
	clockenable();
	setclinttmr(clint, omtime + 20*(timebase/1000000));
	spllo();
	delay(10);
	for (i = 400; ticks == m->ticks && i > 0; i--)
		delay(10);

	splhi();
	if (ticks == m->ticks)
		iprint("clock not interrupting!\n");
	if (clint->mtime <= omtime)
		iprint("clint clock is not advancing.\n");
	return ticks != m->ticks;
}
